from __future__ import annotations

import asyncio
import contextlib
import re
import time
from pathlib import Path
from typing import Any

import discord
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import inline

from pylav.constants.builtin_nodes import BUNDLED_NODES_IDS_HOST_MAPPING
from pylav.core.context import PyLavContext
from pylav.extension.red.ui.buttons.generic import CloseButton, DoneButton, NavigateButton, RefreshButton
from pylav.extension.red.ui.buttons.nodes import (
    NodeButton,
    NodeDeleteButton,
    NodeShowEnabledSourcesButton,
    SearchOnlyNodeToggleButton,
    SSLNodeToggleButton,
)
from pylav.extension.red.ui.menus.generic import BaseMenu
from pylav.extension.red.ui.modals.generic import PromptForInput
from pylav.extension.red.ui.selectors.nodes import NodeSelectSelector, SourceSelector
from pylav.extension.red.ui.sources.nodes import NodeManageSource, NodePickerSource
from pylav.helpers import emojis
from pylav.helpers.format.strings import shorten_string
from pylav.storage.models.node.real import Node
from pylav.type_hints.bot import DISCORD_BOT_TYPE, DISCORD_COG_TYPE, DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

URL_REGEX = re.compile(r"^(https?)://(\S+)$")
_ = Translator("PyLav", Path(__file__))


class AddNodeFlow(discord.ui.View):
    ctx: DISCORD_CONTEXT_TYPE
    message: discord.Message
    author: discord.abc.User

    def __init__(self, cog: DISCORD_COG_TYPE, original_author: discord.abc.User):
        super().__init__(timeout=600)

        self.cog = cog
        self.bot = cog.bot
        self.author = original_author
        self.cancelled = True
        self.completed = asyncio.Event()
        self.done_button = DoneButton(
            style=discord.ButtonStyle.green,
            cog=cog,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            cog=cog,
        )
        self.host_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Enter the domain or IP address of the host")),
            label=shorten_string(max_length=100, string=_("Host")),
            style=discord.TextStyle.short,
            min_length=4,
            max_length=200,
        )
        self.port_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Enter the host port to connect to")),
            label=shorten_string(max_length=100, string=_("Port")),
            style=discord.TextStyle.short,
            min_length=2,
            max_length=5,
        )
        self.password_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Enter the node's password")),
            label=shorten_string(max_length=100, string=_("Password")),
            style=discord.TextStyle.short,
            min_length=1,
            max_length=64,
        )
        self.name_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Enter an easy to know name for the node")),
            label=shorten_string(max_length=100, string=_("Name")),
            style=discord.TextStyle.short,
            min_length=8,
            max_length=64,
        )
        self.resume_timeout_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Enter a timeout in seconds")),
            label=shorten_string(max_length=100, string=_("Timeout")),
            style=discord.TextStyle.short,
            min_length=2,
            max_length=4,
        )
        self.search_only_button = SearchOnlyNodeToggleButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.SEARCH,
        )
        self.ssl_button = SSLNodeToggleButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.SSL,
        )
        self.disabled_sources_selector = SourceSelector(
            cog=self.cog, placeholder=shorten_string(max_length=100, string=_("Source to disable")), row=2
        )
        self.name_button = NodeButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.NAME,
            op="name",
            row=1,
        )
        self.host_button = NodeButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.HOST,
            op="host",
            row=1,
        )
        self.port_button = NodeButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.PORT,
            op="port",
            row=1,
        )
        self.password_button = NodeButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.PASSWORD,
            op="password",
            row=1,
        )
        self.timeout_button = NodeButton(
            cog=self.cog,
            style=discord.ButtonStyle.blurple,
            emoji=emojis.TIMEOUT,
            op="timeout",
            row=1,
        )

        self.name = None
        self.host = None
        self.port = None
        self.password = None
        self.resume_timeout = 600
        self.reconnect_attempts = -1
        self.ssl = False
        self.search_only = False
        self.unique_identifier = int(time.time())
        self.done = False

        self.add_item(self.done_button)
        self.add_item(self.close_button)
        self.add_item(self.search_only_button)
        self.add_item(self.ssl_button)

        self.add_item(self.name_button)
        self.add_item(self.host_button)
        self.add_item(self.port_button)
        self.add_item(self.password_button)
        self.add_item(self.timeout_button)

        self.add_item(self.disabled_sources_selector)
        self.last_interaction = None

    def stop(self):
        super().stop()
        asyncio.ensure_future(self.on_timeout())

    async def on_timeout(self):
        self.completed.set()
        if self.message is None:
            return
        with contextlib.suppress(discord.HTTPException):
            if not self.message.flags.ephemeral:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    async def wait_until_complete(self):
        await asyncio.wait_for(self.completed.wait(), timeout=self.timeout)

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None):
        self.unique_identifier = ctx.message.id
        await self.send_initial_message(ctx, description=description, title=title)

    async def interaction_check(self, interaction: DISCORD_INTERACTION_TYPE):
        """Just extends the default reaction_check to use owner_ids"""
        if (not await self.bot.allowed_by_whitelist_blacklist(interaction.user, guild=interaction.guild)) or (
            self.author and (interaction.user.id != self.author.id)
        ):
            await interaction.response.send_message(
                content=_("You are not authorized to interact with this"), ephemeral=True
            )
            return False
        return True

    async def send_initial_message(
        self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None
    ):
        self.ctx = ctx
        self.message = await ctx.send(
            embed=await self.cog.pylav.construct_embed(description=description, title=title, messageable=ctx),
            view=self,
            ephemeral=True,
        )
        return self.message

    async def prompt_name(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.name_prompt)
        await self.name_prompt.responded.wait()
        self.name = self.name_prompt.response
        await interaction.followup.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Name set to {name}").format(name=inline(self.name)),
                messageable=interaction,
            ),
            ephemeral=True,
        )

    async def prompt_password(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.password_prompt)
        await self.password_prompt.responded.wait()
        self.password = self.password_prompt.response
        await interaction.followup.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Password set to {password}").format(password=inline(self.password)),
                messageable=interaction,
            ),
            ephemeral=True,
        )

    async def prompt_host(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.host_prompt)
        await self.host_prompt.responded.wait()
        if match := URL_REGEX.match(self.host_prompt.response):
            protocol = match.group(0)
            self.ssl = protocol == "https"
            self.host = match.group(1)
        else:
            self.host = self.host_prompt.response
        await interaction.followup.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Host set to {host}").format(host=inline(self.host)),
                messageable=interaction,
            ),
            ephemeral=True,
        )

    async def prompt_port(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.port_prompt)
        await self.port_prompt.responded.wait()
        try:
            self.port = int(self.port_prompt.response)
        except ValueError:
            self.port = None
        if self.port is None:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Invalid port"),
                    messageable=interaction,
                ),
                ephemeral=True,
            )
        else:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Port set to {port}").format(port=inline(f"{self.port}")),
                    messageable=interaction,
                ),
                ephemeral=True,
            )

    async def prompt_resume_timeout(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.resume_timeout_prompt)
        await self.resume_timeout_prompt.responded.wait()
        try:
            self.resume_timeout = int(self.resume_timeout_prompt.response)
        except ValueError:
            self.resume_timeout = None
        if self.resume_timeout is None:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Invalid timeout, it must be a number in seconds"),
                    messageable=interaction,
                ),
                ephemeral=True,
            )
        else:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Timeout set to {timeout} seconds").format(timeout=inline(f"{self.resume_timeout}")),
                    messageable=interaction,
                ),
                ephemeral=True,
            )


class NodePickerMenu(BaseMenu):
    _source: NodePickerSource
    result: Node

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: NodePickerSource,
        selector_text: str,
        selector_cls: type[NodeSelectSelector],  # noqa
        original_author: discord.abc.User,
        *,
        clear_buttons_after: bool = False,
        delete_after_timeout: bool = True,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog,
            bot,
            source,
            clear_buttons_after=clear_buttons_after,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.result: Node | None = None
        self.selector_cls = selector_cls
        self.selector_text = shorten_string(max_length=100, string=selector_text)
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            row=4,
            cog=cog,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            row=4,
            cog=cog,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}",
            direction=0,
            row=4,
            cog=cog,
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            row=4,
            cog=cog,
        )
        self.close_button = CloseButton(
            style=discord.ButtonStyle.red,
            row=4,
            cog=cog,
        )
        self.refresh_button = RefreshButton(
            style=discord.ButtonStyle.grey,
            row=4,
            cog=cog,
        )
        self.select_view: NodeSelectSelector | None = None
        self.author = original_author

    @property
    def source(self) -> NodePickerSource:
        return self._source

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        self.forward_button.disabled = False
        self.backward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        if max_pages == 1:
            self.forward_button.disabled = True
            self.backward_button.disabled = True
            self.first_button.disabled = True
            self.last_button.disabled = True
        elif max_pages == 2:
            self.first_button.disabled = True
            self.last_button.disabled = True
        self.add_item(self.close_button)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        if self.source.select_options:
            options = self.source.select_options
            self.remove_item(self.select_view)
            self.select_view = self.selector_cls(options, self.cog, self.selector_text, self.source.select_mapping)
            self.add_item(self.select_view)
        if self.select_view and not self.source.select_options:
            self.remove_item(self.select_view)
            self.select_view = None

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE):
        if isinstance(ctx, discord.Interaction):
            ctx = await self.cog.bot.get_context(ctx)
        if ctx.interaction and not ctx.interaction.response.is_done():
            await ctx.defer(ephemeral=True)
        self.ctx = ctx
        await self.send_initial_message(ctx)

    async def show_page(self, page_number: int, interaction: DISCORD_INTERACTION_TYPE):
        await self._source.get_page(page_number)
        await self.prepare()
        self.current_page = page_number
        if self.message is not None:
            if not interaction.response.is_done():
                await interaction.response.pong()
            await self.message.edit(view=self)
        elif not interaction.response.is_done():
            await interaction.response.edit_message(view=self)

    async def wait_for_response(self):
        if isinstance(self.select_view, NodeSelectSelector):
            await asyncio.wait_for(self.select_view.responded.wait(), timeout=self.timeout)
            self.result = self.select_view.node


class NodeManagerMenu(BaseMenu):
    _source: NodeManageSource

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        bot: DISCORD_BOT_TYPE,
        source: NodeManageSource,
        original_author: discord.abc.User,
        *,
        delete_after_timeout: bool = True,
        timeout: int = 120,
        message: discord.Message = None,
        starting_page: int = 0,
        **kwargs: Any,
    ) -> None:
        super().__init__(
            cog,
            bot,
            source,
            delete_after_timeout=delete_after_timeout,
            timeout=timeout,
            message=message,
            starting_page=starting_page,
            **kwargs,
        )
        self.unique_identifier = int(time.time())
        self.current_page = -1
        self.forward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=1,
            cog=cog,
            row=0,
        )
        self.backward_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK LEFT-POINTING TRIANGLE}\N{VARIATION SELECTOR-16}",
            direction=-1,
            cog=cog,
            row=0,
        )
        self.first_button = NavigateButton(
            style=discord.ButtonStyle.grey, emoji="\N{BLACK LEFT-POINTING DOUBLE TRIANGLE}", direction=0, cog=cog, row=0
        )
        self.last_button = NavigateButton(
            style=discord.ButtonStyle.grey,
            emoji="\N{BLACK RIGHT-POINTING DOUBLE TRIANGLE}",
            direction=self.source.get_max_pages,
            cog=cog,
            row=0,
        )
        self.close_button = CloseButton(style=discord.ButtonStyle.red, cog=cog, row=0)
        self.host_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Change the domain or IP address of the host")),
            label=shorten_string(max_length=100, string=_("Host")),
            style=discord.TextStyle.short,
            min_length=4,
            max_length=200,
        )
        self.port_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Change the host port to connect to")),
            label=shorten_string(max_length=100, string=_("Port")),
            style=discord.TextStyle.short,
            min_length=2,
            max_length=5,
        )
        self.password_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Change the node's password")),
            label=shorten_string(max_length=100, string=_("Password")),
            style=discord.TextStyle.short,
            min_length=1,
            max_length=64,
        )
        self.name_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Change the name of this node")),
            label=shorten_string(max_length=100, string=_("Name")),
            style=discord.TextStyle.short,
            min_length=8,
            max_length=64,
        )
        self.resume_timeout_prompt = PromptForInput(
            cog=self.cog,
            title=shorten_string(max_length=100, string=_("Enter the new timeout for this node")),
            label=shorten_string(max_length=100, string=_("Timeout")),
            style=discord.TextStyle.short,
            min_length=2,
            max_length=4,
        )
        self.show_sources_button = NodeShowEnabledSourcesButton(cog=self.cog, style=discord.ButtonStyle.blurple, row=1)
        self.done_button = DoneButton(style=discord.ButtonStyle.green, cog=cog, row=1)
        self.search_only_button = SearchOnlyNodeToggleButton(
            cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.SEARCH, row=1
        )
        self.ssl_button = SSLNodeToggleButton(cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.SSL, row=1)

        self.name_button = NodeButton(
            cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.NAME, op="name", row=1
        )
        self.host_button = NodeButton(
            cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.HOST, op="host", row=2
        )
        self.port_button = NodeButton(
            cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.PORT, op="port", row=2
        )
        self.password_button = NodeButton(
            cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.PASSWORD, op="password", row=2
        )
        self.timeout_button = NodeButton(
            cog=self.cog, style=discord.ButtonStyle.blurple, emoji=emojis.TIMEOUT, op="timeout", row=2
        )
        self.delete_button = NodeDeleteButton(cog=self.cog, style=discord.ButtonStyle.red, row=2)
        self.disabled_sources_selector = SourceSelector(
            cog=self.cog, placeholder=shorten_string(max_length=100, string=_("Source to disable")), row=3
        )

        self.cancelled = True

        self.author = original_author
        self.completed = asyncio.Event()

        self.name = None
        self.host = None
        self.port = None
        self.password = None
        self.resume_timeout = None
        self.ssl = None
        self.search_only = None
        self.done = False
        self.delete = None

    @property
    def source(self) -> NodeManageSource:
        return self._source

    async def prepare(self):
        self.clear_items()
        max_pages = self.source.get_max_pages()
        self.done_button.disabled = False
        self.name_button.disabled = False
        self.host_button.disabled = False
        self.port_button.disabled = False
        self.password_button.disabled = False
        self.timeout_button.disabled = False
        self.disabled_sources_selector.disabled = False
        self.ssl_button.disabled = False
        self.search_only_button.disabled = False
        self.backward_button.disabled = False
        self.forward_button.disabled = False
        self.first_button.disabled = False
        self.last_button.disabled = False
        self.add_item(self.close_button)
        self.add_item(self.first_button)
        self.add_item(self.backward_button)
        self.add_item(self.forward_button)
        self.add_item(self.last_button)
        if max_pages <= 2:
            self.first_button.disabled = True
            self.last_button.disabled = True
        if self.source.target:
            self.add_item(self.show_sources_button)
            if self.source.target.identifier not in BUNDLED_NODES_IDS_HOST_MAPPING:
                self.add_item(self.done_button)
                self.add_item(self.search_only_button)
                self.add_item(self.ssl_button)
                self.add_item(self.name_button)
                self.add_item(self.host_button)
                self.add_item(self.port_button)
                self.add_item(self.password_button)
                self.add_item(self.timeout_button)
                self.add_item(self.delete_button)
                self.add_item(self.disabled_sources_selector)

    def stop(self):
        super().stop()
        asyncio.ensure_future(self.on_timeout())

    async def on_timeout(self):
        self.completed.set()
        if self.message is None:
            return
        with contextlib.suppress(discord.HTTPException):
            if not self.message.flags.ephemeral:
                await self.message.delete()
            else:
                await self.message.edit(view=None)

    async def wait_until_complete(self):
        await asyncio.wait_for(self.completed.wait(), timeout=self.timeout)

    async def start(self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None):
        self.unique_identifier = ctx.message.id
        await self.send_initial_message(ctx, description=description, title=title)

    async def interaction_check(self, interaction: DISCORD_INTERACTION_TYPE):
        """Just extends the default reaction_check to use owner_ids"""
        if (not await self.bot.allowed_by_whitelist_blacklist(interaction.user, guild=interaction.guild)) or (
            self.author and (interaction.user.id != self.author.id)
        ):
            await interaction.response.send_message(
                content="You are not authorized to interact with this", ephemeral=True
            )
            return False
        return True

    async def send_initial_message(
        self, ctx: PyLavContext | DISCORD_INTERACTION_TYPE, description: str = None, title: str = None
    ):
        self.ctx = ctx
        await self.prepare()
        self.message = await ctx.send(
            embed=await self.cog.pylav.construct_embed(description=description, title=title, messageable=ctx),
            view=self,
            ephemeral=True,
        )
        return self.message

    async def prompt_name(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.name_prompt)
        await self.name_prompt.responded.wait()
        self.name = self.name_prompt.response
        await interaction.followup.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Name set to {name}").format(name=inline(self.name)),
                messageable=interaction,
            ),
            ephemeral=True,
        )

    async def prompt_password(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.password_prompt)
        await self.password_prompt.responded.wait()
        self.password = self.password_prompt.response
        await interaction.followup.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Password set to {password}").format(password=inline(self.password)),
                messageable=interaction,
            ),
            ephemeral=True,
        )

    async def prompt_host(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.host_prompt)
        await self.host_prompt.responded.wait()
        if match := URL_REGEX.match(self.host_prompt.response):
            protocol = match.group(0)
            self.ssl = protocol == "https"
            self.host = match.group(1)
        else:
            self.host = self.host_prompt.response
        await interaction.followup.send(
            embed=await self.cog.pylav.construct_embed(
                description=_("Host set to {host}").format(host=inline(self.host)),
                messageable=interaction,
            ),
            ephemeral=True,
        )

    async def prompt_port(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.port_prompt)
        await self.port_prompt.responded.wait()
        try:
            self.port = int(self.port_prompt.response)
        except ValueError:
            self.port = None
        if self.port is None:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Invalid port"),
                    messageable=interaction,
                ),
                ephemeral=True,
            )
        else:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Port set to {port}").format(port=inline(f"{self.port}")),
                    messageable=interaction,
                ),
                ephemeral=True,
            )

    async def prompt_resume_timeout(self, interaction: DISCORD_INTERACTION_TYPE) -> None:
        self.cancelled = False
        await interaction.response.send_modal(self.resume_timeout_prompt)
        await self.resume_timeout_prompt.responded.wait()
        try:
            self.resume_timeout = int(self.resume_timeout_prompt.response)
        except ValueError:
            self.resume_timeout = None
        if self.resume_timeout is None:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Invalid timeout, it must be a number in seconds"),
                    messageable=interaction,
                ),
                ephemeral=True,
            )
        else:
            await interaction.followup.send(
                embed=await self.cog.pylav.construct_embed(
                    description=_("Timeout set to {timeout} seconds").format(timeout=inline(f"{self.resume_timeout}")),
                    messageable=interaction,
                ),
                ephemeral=True,
            )



from __future__ import annotations

import asyncio
import socket
from math import atan2, cos, radians, sin, sqrt

import aiohttp
import asyncstdlib
import ujson

from pylav.constants.coordinates import REGION_TO_COUNTRY_COORDINATE_MAPPING


def distance(lat1: float, lon1: float, lat2: float, lon2: float) -> float:
    dlat = radians(lat2) - radians(lat1)
    dlon = radians(lon2) - radians(lon1)
    a = sin(dlat / 2) * sin(dlat / 2) + cos(radians(lat1)) * cos(radians(lat2)) * sin(dlon / 2) * sin(dlon / 2)
    return 6371 * 2 * atan2(sqrt(a), sqrt(1 - a))


async def closest(
    data: dict[str, tuple[float, float]], compare_to: tuple[float, float], *, region_pool: set[str] | None = None
) -> tuple[str, tuple[float, float]]:
    if region_pool is None:
        region_pool = set()
    entries = [(k, v) for k, v in data.items() if not region_pool or k in region_pool]
    return await asyncstdlib.min(
        entries,
        key=lambda p: distance(compare_to[0], compare_to[1], p[1][0], p[1][1]),
    )


async def get_closest_region_name_and_coordinate(
    lat: float, lon: float, region_pool: set[str] | None = None
) -> tuple[str, tuple[float, float]]:
    closest_region, closest_coordinates = await closest(
        REGION_TO_COUNTRY_COORDINATE_MAPPING, compare_to=(lat, lon), region_pool=region_pool
    )
    name, coordinate = await asyncstdlib.anext(
        asyncstdlib.iter((k, v) for k, v in REGION_TO_COUNTRY_COORDINATE_MAPPING.items() if v == closest_coordinates)
    )
    return name, coordinate


async def get_coordinates(ip: str | None = None) -> tuple[float, ...]:
    url = f"https://ipinfo.io/{ip}/json" if ip else "https://ipinfo.io/json"
    async with aiohttp.ClientSession(json_serialize=ujson.dumps) as session:
        async with session.get(url) as response:
            data = await response.json(loads=ujson.loads)
            return tuple(map(float, data["loc"].split(",")))


async def get_closest_discord_region(host: str | None = None) -> tuple[str, tuple[float, float]]:
    try:
        if host is None or host in ["localhost", "127.0.0.1", "::1", "0.0.0.0", "::"]:
            host_ip = None
        else:
            host_ip = await asyncio.to_thread(socket.gethostbyname, host)
    except Exception:  # noqa
        host_ip = None  # If there's any issues getting the ip from the hostname, just use the host ip
    try:
        latitude, longitude = await get_coordinates(host_ip)
        return await get_closest_region_name_and_coordinate(lat=latitude, lon=longitude)
    except Exception:  # noqa
        return "unknown_pylav", (0, 0)



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING, TypeVar

import asyncstdlib
from asyncstdlib import heapq
from discord.app_commands import Choice, Transformer
from discord.ext import commands
from rapidfuzz import fuzz

from pylav.exceptions.database import EntryNotFoundException
from pylav.extension.radio.objects import Codec, Country, CountryCode, Language, State, Station, Tag
from pylav.helpers.format.strings import shorten_string
from pylav.logging import getLogger
from pylav.type_hints.bot import DISCORD_CONTEXT_TYPE, DISCORD_INTERACTION_TYPE

try:
    from redbot.core.i18n import Translator  # type: ignore

    _ = Translator("PyLav", Path(__file__))
except ImportError:
    Translator = None

    def _(string: str) -> str:
        return string


LOGGER = getLogger("PyLav.extension.Shared.converters.radio")

if TYPE_CHECKING:

    StationConverter = TypeVar("StationConverter", bound=list[Station])
    TagConverter = TypeVar("TagConverter", bound=list[Tag])
    LanguageConverter = TypeVar("LanguageConverter", bound=list[Language])
    StateConverter = TypeVar("StateConverter", bound=list[State])
    CodecConverter = TypeVar("CodecConverter", bound=list[Codec])
    CountryCodeConverter = TypeVar("CountryCodeConverter", bound=list[CountryCode])
    CountryConverter = TypeVar("CountryConverter", bound=list[Country])
else:

    class StationConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Station]:
            """Converts a station name to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                return await ctx.pylav.radio_browser.station_by_uuid(
                    arg
                ) or await ctx.pylav.radio_browser.stations_by_name(name=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Station with name `{arg}` not found").format(arg=arg)) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Station]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            data = interaction.data
            options = data.get("options", [])
            kwargs = {"order": "votes"}
            if options:
                country_code = [v for v in options if v.get("name") == "countrycode"]
                country = [v for v in options if v.get("name") == "country"]
                state = [v for v in options if v.get("name") == "state"]
                language = [v for v in options if v.get("name") == "language"]
                tags = [v for v in options if v.get("name", "").startswith("tag")]
                if country_code and (val := country_code[0].get("value")):
                    kwargs["countrycode"] = val
                if country and (val := country[0].get("value")):
                    kwargs["country"] = val
                    kwargs["country_exact"] = any(
                        True for c in await interaction.client.pylav.radio_browser.countries() if c.name == val
                    )
                if state and (val := state[0].get("value")):
                    kwargs["state"] = val
                    kwargs["state_exact"] = any(
                        True for t in await interaction.client.pylav.radio_browser.states() if t.name == val
                    )
                if language and (val := language[0].get("value")):
                    kwargs["language"] = val
                    kwargs["language_exact"] = any(
                        True for t in await interaction.client.pylav.radio_browser.languages() if t.name == val
                    )
                if tags:
                    if len(tags) == 1 and (val := tags[0].get("value")):
                        kwargs["tag"] = val
                        kwargs["tag_exact"] = any(
                            True for t in await interaction.client.pylav.radio_browser.tags() if t.name == val
                        )
                    elif len(tags) > 1:
                        kwargs["tag_list"] = ",".join([tv for t in tags if (tv := t.get("value"))])

            if current:
                kwargs["name"] = current
            if not current:
                return [
                    Choice(
                        name=shorten_string(e.name, max_length=100)
                        if e.name
                        else shorten_string(max_length=100, string=_("Unnamed")),
                        value=f"{e.stationuuid}",
                    )
                    for e in await interaction.client.pylav.radio_browser.stations_by_votes(limit=25)
                ]
            if not kwargs:
                kwargs["limit"] = 1000
            stations = await interaction.client.pylav.radio_browser.search(**kwargs)

            async def _filter(c: Station):
                return (
                    await asyncio.to_thread(
                        fuzz.partial_ratio,
                        c.name,
                        current,
                    ),
                    c.votes,
                )

            extracted: list[Station] = await heapq.nlargest(asyncstdlib.iter(stations), n=25, key=_filter)

            return [
                Choice(
                    name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.stationuuid}"
                )
                for e in extracted
            ]

    class TagConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Tag]:
            """Converts a Tag name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                return await ctx.pylav.radio_browser.tags(arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Tag with name `{arg}` not found").format(arg=arg)) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Tag]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            tags = await interaction.client.pylav.radio_browser.tags()
            if not current:
                return [
                    Choice(
                        name=shorten_string(e.name, max_length=100)
                        if e.name
                        else shorten_string(max_length=100, string=_("Unnamed")),
                        value=f"{e.name}",
                    )
                    for e in tags
                ][:25]

            async def _filter(c):
                return await asyncio.to_thread(fuzz.partial_ratio, c.name, current), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(tags), n=25, key=_filter)

            return [
                Choice(
                    name=shorten_string(e.name, max_length=100)
                    if e.name
                    else shorten_string(max_length=100, string=_("Unnamed")),
                    value=f"{e.name}",
                )
                for e in extracted
            ]

    class LanguageConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Language]:
            """Converts a Language name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                return await ctx.pylav.radio_browser.languages(arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Language with name `{arg}` not found").format(arg=arg)) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Language]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            languages = await interaction.client.pylav.radio_browser.languages()
            if not current:
                return [
                    Choice(
                        name=shorten_string(e.name, max_length=100)
                        if e.name
                        else shorten_string(max_length=100, string=_("Unnamed")),
                        value=f"{e.name}",
                    )
                    for e in languages
                ][:25]

            async def _filter(c):
                return await asyncio.to_thread(fuzz.partial_ratio, c.name, current), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(languages), n=25, key=_filter)

            return [
                Choice(
                    name=shorten_string(e.name, max_length=100)
                    if e.name
                    else shorten_string(max_length=100, string=_("Unnamed")),
                    value=f"{e.name}",
                )
                for e in extracted
            ]

    class StateConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[State]:
            """Converts a State name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []
            try:
                return await ctx.pylav.radio_browser.states(state=arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("State with name `{arg}` not found")) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[State]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            data = interaction.data
            options = data.get("options", [])
            kwargs = {}
            if options:
                country = [v for v in options if v.get("name") == "country"]

                if country and (val := country[0].get("value")):
                    kwargs["country"] = val
            states = await interaction.client.pylav.radio_browser.states(**kwargs)
            if not current:
                return [
                    Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                    for e in states
                ][:25]

            async def _filter(c):
                return await asyncio.to_thread(fuzz.partial_ratio, c.name, current), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(states), n=25, key=_filter)

            return [
                Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                for e in extracted
            ]

    class CodecConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Codec]:
            """Converts a Codec name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                return await ctx.pylav.radio_browser.codecs(arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Codec with name `{arg}` not found").format(arg=arg)) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Codec]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            codecs = await interaction.client.pylav.radio_browser.codecs()
            if not current:
                return [
                    Choice(
                        name=shorten_string(e.name, max_length=100)
                        if e.name
                        else shorten_string(max_length=100, string=_("Unnamed")),
                        value=f"{e.name}",
                    )
                    for e in codecs
                ][:25]

            async def _filter(c):
                return await asyncio.to_thread(fuzz.ratio, c.name, current), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(codecs), n=25, key=_filter)

            return [
                Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                for e in extracted
            ]

    class CountryCodeConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[CountryCode]:
            """Converts a CountryCode name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                return await ctx.pylav.radio_browser.countrycodes(arg)
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Country code `{arg}` not found").format(arg=arg)) from e

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[CountryCode]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            country_codes = await interaction.client.pylav.radio_browser.countrycodes()
            if not current:
                return [
                    Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                    for e in country_codes
                ][:25]

            async def _filter(c):
                return await asyncio.to_thread(fuzz.partial_ratio, c.name, current), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(country_codes), n=25, key=_filter)

            return [
                Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                for e in extracted
            ]

    class CountryConverter(Transformer):
        @classmethod
        async def convert(cls, ctx: DISCORD_CONTEXT_TYPE, arg: str) -> list[Country]:
            """Converts a Country name to to a matching object"""
            if ctx.bot.pylav.radio_browser.disabled:
                return []

            try:
                countries = await ctx.pylav.radio_browser.countries()
            except EntryNotFoundException as e:
                raise commands.BadArgument(_("Country with name `{arg}` not found").format(arg=arg)) from e

            async def _filter(c):
                return await asyncio.to_thread(fuzz.partial_ratio, arg, c.name), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(countries), n=25, key=_filter)
            return extracted

        @classmethod
        async def transform(cls, interaction: DISCORD_INTERACTION_TYPE, argument: str) -> list[Country]:
            if not interaction.response.is_done():
                await interaction.response.defer(ephemeral=True)
            ctx = await interaction.client.get_context(interaction)
            if interaction.client.pylav.radio_browser.disabled:
                return []
            return await cls.convert(ctx, argument)

        @classmethod
        async def autocomplete(cls, interaction: DISCORD_INTERACTION_TYPE, current: str) -> list[Choice]:
            if interaction.client.pylav.radio_browser.disabled:
                return [
                    Choice(
                        name=shorten_string(max_length=100, string=_("Radio Browser is disabled")),
                        value="???",
                    )
                ]
            data = interaction.data
            options = data.get("options", [])
            kwargs = {}
            if options:
                code = [v for v in options if v.get("name") in ["countrycode", "code"]]
                if code and (val := code[0].get("value")):
                    kwargs["code"] = val
            countries = await interaction.client.pylav.radio_browser.countries(**kwargs)
            if not current:
                return [
                    Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                    for e in countries
                ][:25]

            async def _filter(c):
                return await asyncio.to_thread(fuzz.partial_ratio, c.name, current), [-ord(c) for c in c.name]

            extracted = await heapq.nlargest(asyncstdlib.iter(countries), n=25, key=_filter)

            return [
                Choice(name=shorten_string(e.name, max_length=100) if e.name else _("Unnamed"), value=f"{e.name}")
                for e in extracted
            ]



from __future__ import annotations

import contextlib
from pathlib import Path
from typing import Literal

import discord
from redbot.core.i18n import Translator

from pylav.helpers import emojis
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE
from pylav.utils.vendor.redbot import AsyncIter

_ = Translator("PyLav", Path(__file__))


class DisconnectButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emojis.POWER,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not await self.view.bot.is_owner(context.author):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, title=_("You are not authorized to perform this action")
                ),
                ephemeral=True,
            )
            return
        player = self.view.source.current_player
        if not player:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, title=_("No Player Available For Action - Try Refreshing")
                ),
                ephemeral=True,
            )
            return
        if notify_channel := await player.notify_channel():
            with contextlib.suppress(discord.HTTPException):
                await notify_channel.send(
                    embed=await self.cog.pylav.construct_embed(
                        title=_("Bot Owner Action"), description=_("Player disconnected")
                    )
                )
        await player.disconnect(requester=context.author)

        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        await self.view.message.edit(view=self.view, **kwargs)


class StopTrackButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        style: discord.ButtonStyle,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emojis.STOP,
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not await self.view.bot.is_owner(context.author):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, description=_("You are not authorized to perform this action")
                ),
                ephemeral=True,
            )
            return
        player = self.view.source.current_player
        if not player:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, description=_("No Player Available For Action - Try Refreshing")
                ),
                ephemeral=True,
            )
            return

        await player.stop()
        if notify_channel := await player.notify_channel():
            with contextlib.suppress(discord.HTTPException):
                await notify_channel.send(
                    embed=await self.cog.pylav.construct_embed(
                        title=_("Bot Owner Action"), description=_("Player stopped")
                    )
                )

        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        await self.view.message.edit(view=self.view, **kwargs)


class DisconnectAllButton(discord.ui.Button):
    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        disconnect_type: Literal["all", "inactive"],
        style: discord.ButtonStyle,
        row: int = None,
    ):
        super().__init__(
            style=style,
            emoji=emojis.POWER,
            row=row,
        )

        self.disconnect_type = disconnect_type
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if not interaction.response.is_done():
            await interaction.response.defer(ephemeral=True)
        context = await self.cog.bot.get_context(interaction)
        if not await self.view.bot.is_owner(context.author):
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, description=_("You are not authorized to perform this action")
                ),
                ephemeral=True,
            )
            return

        players = (
            self.cog.pylav.player_manager.connected_players
            if self.disconnect_type == "all"
            else self.cog.pylav.player_manager.not_playing_players
        )
        if not players:
            await context.send(
                embed=await self.cog.pylav.construct_embed(
                    messageable=context, description=_("No Players Available For Action - Try Refreshing")
                ),
                ephemeral=True,
            )
            return
        async for player in AsyncIter(players):
            if notify_channel := await player.notify_channel():
                with contextlib.suppress(discord.HTTPException):
                    await notify_channel.send(
                        embed=await self.cog.pylav.construct_embed(
                            title=_("Bot Owner Action"), description=_("Player disconnected")
                        )
                    )
            await player.disconnect(requester=context.author)

        await self.view.prepare()
        kwargs = await self.view.get_page(self.view.current_page)
        await self.view.message.edit(view=self.view, **kwargs)



from __future__ import annotations

import discord

from pylav.helpers.format.strings import shorten_string
from pylav.type_hints.generics import ANY_GENERIC_TYPE


class EntryOption(discord.SelectOption):
    @classmethod
    async def from_entry(cls, entry: ANY_GENERIC_TYPE, index: int):
        return cls(
            label=shorten_string(max_length=100, string=f"{index + 1}. {entry.name}"),
            value=f"{entry.id}",
        )



from __future__ import annotations

import asyncio
import typing

from dacite import from_dict  # type: ignore

from pylav.logging import getLogger
from pylav.nodes.api.responses.track import Track
from pylav.utils.vendor.lavalink_py.datarw import DataReader

LOGGER = getLogger("PyLav.Track.Decoder")


# noinspection SpellCheckingInspection,PyPep8Naming
def decode_track(track: str) -> Track:
    """Decodes a base64 track string into a Track object.

    Parameters
    ----------
    track: :class:`str`
        The base64 track string.

    Returns
    -------
    :class:`Track`
    The decoded Track object
    """
    reader = DataReader(track)
    version = reader.read_version()
    plugin_info = {}
    title = reader.read_utfm()
    author = reader.read_utfm()
    length = reader.read_long()
    identifier = reader.read_utf()
    is_stream = reader.read_boolean()
    uri = reader.read_nullable_utf() if version >= 2 else None
    if version >= 3:
        artworkUrl = reader.read_nullable_utf()
        isrc = reader.read_nullable_utf()
    else:
        artworkUrl = None
        isrc = None
    source = reader.read_utf()
    try:
        match source:
            case "local" | "http" if version >= 2:
                plugin_info["probeInfo"] = reader.read_utfm()
            case "spotify" | "applemusic" | "deezer" if version == 2:
                isrc = reader.read_nullable_utf()
                artworkUrl = reader.read_nullable_utf()
            case "yandexmusic" if version == 2:
                artworkUrl = reader.read_nullable_utf()
        # Position
        __ = reader.read_long()  # Discard position, we don't need it
    except Exception as exc:
        LOGGER.verbose("Error while decoding version %d track: %s", version, track, exc_info=exc)

    return typing.cast(
        Track,
        from_dict(
            data_class=Track,
            data={
                "encoded": track,
                "info": {
                    "version": version,
                    "title": title,
                    "author": author,
                    "length": length,
                    "identifier": identifier,
                    "isStream": is_stream,
                    "uri": uri,
                    "isSeekable": not is_stream,
                    "sourceName": source,
                    "artworkUrl": artworkUrl,
                    "isrc": isrc,
                    "position": 0,
                },
                "pluginInfo": plugin_info,
            },
        ),
    )


async def async_decoder(track: str) -> Track:
    return await asyncio.to_thread(decode_track, track=track)



from __future__ import annotations

import logging
import os

from red_commons.logging import RedTraceLogger  # type: ignore
from red_commons.logging import getLogger as redgetLogger
from red_commons.logging import maybe_update_logger_class

__all__ = ("getLogger",)

maybe_update_logger_class()

LOGGER_PREFIX = os.getenv("PYLAV__LOGGER_PREFIX", "")

logging.getLogger("deepdiff.diff").setLevel(logging.FATAL)
logging.getLogger("apscheduler").setLevel(logging.ERROR)
logging.getLogger("aiohttp_client_cache").setLevel(logging.ERROR)


# noinspection PyPep8Naming
def getLogger(name: str) -> RedTraceLogger:  # noqa: N802
    return redgetLogger(f"{LOGGER_PREFIX}{name}")



from __future__ import annotations

import os

from pylav.extension.m3u.parser import is_url, urljoin


def _urijoin(base_uri: str, path: str) -> str:
    if not is_url(base_uri):
        return os.path.normpath(os.path.join(base_uri, path.strip("/")))
    if base_uri[-1] != "/":
        base_uri += "/"
    return urljoin(base_uri, path)


class BasePathMixin:
    __slots__ = ("base_uri", "uri")
    base_uri: str | None
    uri: str | None

    @property
    def absolute_uri(self) -> str | None:
        if self.uri is None:
            return None
        if is_url(self.uri):
            return self.uri
        if self.base_uri is None:
            raise ValueError("There can not be `absolute_uri` with no `base_uri` set")
        return _urijoin(self.base_uri, self.uri)

    @property
    def base_path(self) -> str | None:
        return None if self.uri is None else os.path.dirname(self.get_path_from_uri())

    def get_path_from_uri(self) -> str:
        """Some URIs have a slash in the query string"""
        return self.uri.split("?")[0]

    @base_path.setter
    def base_path(self, newbase_path: str) -> None:
        if self.uri is not None:
            self.uri = (
                self.uri.replace(self.base_path, newbase_path) if self.base_path else f"{newbase_path}/{self.uri}"
            )


class GroupedBasePathMixin:
    __slots__ = ()

    def _set_base_uri(self, new_base_uri: str) -> None:
        for item in self:  # type: ignore
            item.base_uri = new_base_uri

    base_uri = property(None, _set_base_uri)

    def _set_base_path(self, newbase_path: str) -> None:
        for item in self:  # type: ignore
            item.base_path = newbase_path

    base_path = property(None, _set_base_path)



from __future__ import annotations

import asyncio
from pathlib import Path
from typing import TYPE_CHECKING

import discord
from redbot.core.i18n import Translator
from redbot.core.utils.chat_formatting import humanize_list

from pylav.constants.node_features import SUPPORTED_FEATURES, SUPPORTED_SOURCES
from pylav.extension.red.ui.selectors.options.nodes import SOURCE_OPTIONS, NodeOption
from pylav.helpers.format.strings import shorten_string
from pylav.nodes.node import Node
from pylav.storage.models.node.real import Node as NodeModel
from pylav.type_hints.bot import DISCORD_COG_TYPE, DISCORD_INTERACTION_TYPE

if TYPE_CHECKING:
    from pylav.extension.red.ui.menus.nodes import AddNodeFlow

_ = Translator("PyLav", Path(__file__))


class SourceSelector(discord.ui.Select):
    view: AddNodeFlow

    def __init__(
        self,
        cog: DISCORD_COG_TYPE,
        row: int | None = None,
        placeholder: str = "",
    ):
        super().__init__(
            min_values=1,
            max_values=len(SUPPORTED_SOURCES.union(SUPPORTED_FEATURES)),
            options=SOURCE_OPTIONS,
            placeholder=shorten_string(max_length=100, string=placeholder),
            row=row,
        )
        self.cog = cog

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        if self.view.author.id != interaction.user.id:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(
                    messageable=interaction, description=_("You are not authorized to interact with this option")
                ),
                ephemeral=True,
            )
        await interaction.response.send_message(
            embed=await self.cog.pylav.construct_embed(
                messageable=interaction,
                description=_("Disabling the following sources: {sources}").format(sources=humanize_list(self.values)),
            ),
            ephemeral=True,
        )


class NodeSelectSelector(discord.ui.Select):
    def __init__(
        self,
        options: list[NodeOption],
        cog: DISCORD_COG_TYPE,
        placeholder: str,
        mapping: dict[str, Node],
    ):
        super().__init__(
            min_values=1,
            max_values=1,
            options=options,
            placeholder=shorten_string(max_length=100, string=placeholder),
        )
        self.cog = cog
        self.mapping = mapping
        self.node: NodeModel = None  # type:ignore
        self.responded = asyncio.Event()

    async def callback(self, interaction: DISCORD_INTERACTION_TYPE):
        playlist_id = self.values[0]
        self.node: Node = self.mapping.get(playlist_id)
        if self.node is None:
            await interaction.response.send_message(
                embed=await self.cog.pylav.construct_embed(messageable=interaction, title=_("Node not found")),
                ephemeral=True,
            )
            self.view.stop()
            await self.view.on_timeout()
            return
        self.responded.set()
        self.view.stop()
        await self.view.on_timeout()



from __future__ import annotations

import base64
import os
from copy import deepcopy
from typing import Any, cast

import yaml
from deepdiff import DeepDiff  # type: ignore

# noinspection PyProtectedMember
from pylav._internals.functions import _get_path
from pylav.constants.config import ENV_FILE
from pylav.constants.config.utils import remove_keys
from pylav.constants.node_features import SUPPORTED_SEARCHES
from pylav.constants.specials import ANIME
from pylav.logging import getLogger

LOGGER = getLogger("PyLav.Environment")

data = cast(dict[str, Any], yaml.safe_load(ENV_FILE.open(mode="r").read()))
data_new = deepcopy(data)

if (POSTGRES_PORT := data.get("PYLAV__POSTGRES_PORT")) is None:
    # noinspection SpellCheckingInspection
    POSTGRES_PORT = os.getenv("PYLAV__POSTGRES_PORT", os.getenv("PGPORT"))
    data_new["PYLAV__POSTGRES_PORT"] = POSTGRES_PORT

if (POSTGRES_PASSWORD := data.get("PYLAV__POSTGRES_PASSWORD")) is None:
    # noinspection SpellCheckingInspection
    POSTGRES_PASSWORD = os.getenv("PYLAV__POSTGRES_PASSWORD", os.getenv("PGPASSWORD"))
    data_new["PYLAV__POSTGRES_PASSWORD"] = POSTGRES_PASSWORD

if (POSTGRES_USER := data.get("PYLAV__POSTGRES_USER")) is None:
    # noinspection SpellCheckingInspection
    POSTGRES_USER = os.getenv("PYLAV__POSTGRES_USER", os.getenv("PGUSER"))
    data_new["PYLAV__POSTGRES_USER"] = POSTGRES_USER

if (POSTGRES_DATABASE := data.get("PYLAV__POSTGRES_DB")) is None:
    # noinspection SpellCheckingInspection
    POSTGRES_DATABASE = os.getenv("PYLAV__POSTGRES_DB", os.getenv("PGDATABASE"))
    data_new["PYLAV__POSTGRES_DB"] = POSTGRES_DATABASE

if (POSTGRES_HOST := data.get("PYLAV__POSTGRES_HOST")) is None:
    POSTGRES_HOST = os.getenv("PYLAV__POSTGRES_HOST", os.getenv("PGHOST"))
    data_new["PYLAV__POSTGRES_HOST"] = POSTGRES_HOST

if (POSTGRES_SOCKET := data.get("PYLAV__POSTGRES_SOCKET")) is None:
    POSTGRES_SOCKET = os.getenv("PYLAV__POSTGRES_SOCKET")
    data_new["PYLAV__POSTGRES_SOCKET"] = POSTGRES_SOCKET

FALLBACK_POSTGREST_HOST = POSTGRES_HOST
if POSTGRES_SOCKET is not None:
    POSTGRES_PORT = None
    POSTGRES_HOST = POSTGRES_SOCKET

if (REDIS_FULL_ADDRESS_RESPONSE_CACHE := data.get("PYLAV__REDIS_FULL_ADDRESS_RESPONSE_CACHE")) is None:
    REDIS_FULL_ADDRESS_RESPONSE_CACHE = os.getenv("PYLAV__REDIS_FULL_ADDRESS_RESPONSE_CACHE")
    data_new["PYLAV__REDIS_FULL_ADDRESS_RESPONSE_CACHE"] = REDIS_FULL_ADDRESS_RESPONSE_CACHE

if (LINKED_BOT_IDS := data.get("PYLAV__LINKED_BOT_IDS")) is None:
    LINKED_BOT_IDS = list(map(str.strip, os.getenv("PYLAV__LINKED_BOT_IDS", "").split("|")))
    data_new["PYLAV__LINKED_BOT_IDS"] = LINKED_BOT_IDS

if (USE_BUNDLED_EXTERNAL_PYLAV_NODE := data.get("PYLAV__USE_BUNDLED_EXTERNAL_PYLAV_NODE")) is None:
    USE_BUNDLED_EXTERNAL_PYLAV_NODE = bool(int(os.getenv("PYLAV__USE_BUNDLED_EXTERNAL_PYLAV_NODE", "0")))
    data_new["PYLAV__USE_BUNDLED_EXTERNAL_PYLAV_NODE"] = USE_BUNDLED_EXTERNAL_PYLAV_NODE

if (READ_CACHING_ENABLED := data.get("PYLAV__READ_CACHING_ENABLED")) is None:
    READ_CACHING_ENABLED = bool(int(os.getenv("PYLAV__READ_CACHING_ENABLED", "0")))
    data_new["PYLAV__READ_CACHING_ENABLED"] = READ_CACHING_ENABLED
if (JAVA_EXECUTABLE := data.get("PYLAV__JAVA_EXECUTABLE")) is None:
    JAVA_EXECUTABLE = _get_path(os.getenv("PYLAV__JAVA_EXECUTABLE") or "java")
    data_new["PYLAV__JAVA_EXECUTABLE"] = JAVA_EXECUTABLE

if (EXTERNAL_UNMANAGED_HOST := data.get("PYLAV__EXTERNAL_UNMANAGED_HOST")) is None:
    EXTERNAL_UNMANAGED_HOST = os.getenv("PYLAV__EXTERNAL_UNMANAGED_HOST")
    data_new["PYLAV__EXTERNAL_UNMANAGED_HOST"] = EXTERNAL_UNMANAGED_HOST

if (EXTERNAL_UNMANAGED_PORT := data.get("PYLAV__EXTERNAL_UNMANAGED_PORT")) is None:
    EXTERNAL_UNMANAGED_PORT = int(os.getenv("PYLAV__EXTERNAL_UNMANAGED_PORT", "80"))
    data_new["PYLAV__EXTERNAL_UNMANAGED_PORT"] = EXTERNAL_UNMANAGED_PORT

if (EXTERNAL_UNMANAGED_PASSWORD := data.get("PYLAV__EXTERNAL_UNMANAGED_PASSWORD")) is None:
    EXTERNAL_UNMANAGED_PASSWORD = os.getenv("PYLAV__EXTERNAL_UNMANAGED_PASSWORD")
    data_new["PYLAV__EXTERNAL_UNMANAGED_PASSWORD"] = EXTERNAL_UNMANAGED_PASSWORD

if (EXTERNAL_UNMANAGED_SSL := data.get("PYLAV__EXTERNAL_UNMANAGED_SSL")) is None:
    EXTERNAL_UNMANAGED_SSL = bool(int(os.getenv("PYLAV__EXTERNAL_UNMANAGED_SSL", "0")))
    data_new["PYLAV__EXTERNAL_UNMANAGED_SSL"] = EXTERNAL_UNMANAGED_SSL

if (TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS := data.get("PYLAV__TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS")) is None:
    TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS = max(
        int(os.getenv("PYLAV__TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS", "1")), 1
    )
    data_new["PYLAV__TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS"] = TASK_TIMER_UPDATE_BUNDLED_PLAYLISTS_DAYS

if (
    TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS := data.get(
        "PYLAV__TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS"
    )
) is None:
    TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS = max(
        int(os.getenv("PYLAV__TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS", "7")), 7
    )
    data_new[
        "PYLAV__TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS"
    ] = TASK_TIMER_UPDATE_BUNDLED_EXTERNAL_PLAYLISTS_DAYS

if (TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS := data.get("PYLAV__TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS")) is None:
    TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS = max(
        int(os.getenv("PYLAV__TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS", "7")), 7
    )
    data_new["PYLAV__TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS"] = TASK_TIMER_UPDATE_EXTERNAL_PLAYLISTS_DAYS

if (DEFAULT_SEARCH_SOURCE := data.get("PYLAV__DEFAULT_SEARCH_SOURCE")) is None:
    DEFAULT_SEARCH_SOURCE = os.getenv("PYLAV__DEFAULT_SEARCH_SOURCE")

if DEFAULT_SEARCH_SOURCE not in SUPPORTED_SEARCHES:
    # noinspection SpellCheckingInspection
    LOGGER.warning("Invalid search source %s, defaulting to dzsearch", DEFAULT_SEARCH_SOURCE)
    LOGGER.info("Valid search sources are %s", ", ".join(SUPPORTED_SEARCHES.keys()))
    # noinspection SpellCheckingInspection
    DEFAULT_SEARCH_SOURCE = "dzsearch"
data_new["PYLAV__DEFAULT_SEARCH_SOURCE"] = DEFAULT_SEARCH_SOURCE

if (MANAGED_NODE_SPOTIFY_CLIENT_ID := data.get("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_ID")) is None:
    MANAGED_NODE_SPOTIFY_CLIENT_ID = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_ID", "")
    data_new["PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_ID"] = MANAGED_NODE_SPOTIFY_CLIENT_ID

if (MANAGED_NODE_SPOTIFY_CLIENT_SECRET := data.get("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_SECRET")) is None:
    MANAGED_NODE_SPOTIFY_CLIENT_SECRET = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_SECRET") or ""
    data_new["PYLAV__MANAGED_NODE_SPOTIFY_CLIENT_SECRET"] = MANAGED_NODE_SPOTIFY_CLIENT_SECRET

if (MANAGED_NODE_SPOTIFY_COUNTRY_CODE := data.get("PYLAV__MANAGED_NODE_SPOTIFY_COUNTRY_CODE")) is None:
    MANAGED_NODE_SPOTIFY_COUNTRY_CODE = os.getenv("PYLAV__MANAGED_NODE_SPOTIFY_COUNTRY_CODE") or "US"
    data_new["PYLAV__MANAGED_NODE_SPOTIFY_COUNTRY_CODE"] = MANAGED_NODE_SPOTIFY_COUNTRY_CODE

if (MANAGED_NODE_APPLE_MUSIC_API_KEY := data.get("PYLAV__MANAGED_NODE_APPLE_MUSIC_API_KEY")) is None:
    MANAGED_NODE_APPLE_MUSIC_API_KEY = os.getenv("PYLAV__MANAGED_NODE_APPLE_MUSIC_API_KEY") or ""
    data_new["PYLAV__MANAGED_NODE_APPLE_MUSIC_API_KEY"] = MANAGED_NODE_APPLE_MUSIC_API_KEY

if (MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE := data.get("PYLAV__MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE")) is None:
    MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE = os.getenv("PYLAV__MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE") or "US"
    data_new["PYLAV__MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE"] = MANAGED_NODE_APPLE_MUSIC_COUNTRY_CODE

if (MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN := data.get("PYLAV__MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN")) is None:
    MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN = os.getenv("PYLAV__MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN") or ""
    data_new["PYLAV__MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN"] = MANAGED_NODE_YANDEX_MUSIC_ACCESS_TOKEN

if (MANAGED_NODE_DEEZER_KEY := data.get("PYLAV__MANAGED_NODE_DEEZER_KEY")) is None:
    MANAGED_NODE_DEEZER_KEY = os.getenv("PYLAV__MANAGED_NODE_DEEZER_KEY") or "".join(
        [base64.b64decode(r).decode() for r in ANIME.split(b"|")]
    )
    data_new["PYLAV__MANAGED_NODE_DEEZER_KEY"] = MANAGED_NODE_DEEZER_KEY

data_new = remove_keys("PYLAV__CACHING_ENABLED", data=data_new)

if DeepDiff(data, data_new, ignore_order=True, max_passes=2, cache_size=1000):
    with ENV_FILE.open(mode="w") as file:
        LOGGER.info("Updating %s with the following content: %r", ENV_FILE, data_new)
        yaml.safe_dump(data_new, file, default_flow_style=False, sort_keys=False, encoding="utf-8")



from __future__ import annotations

from pylav.exceptions.base import PyLavException


class PlaylistException(PyLavException):
    """Base class for playlist related errors"""


class InvalidPlaylistException(PlaylistException):
    """Raised when a playlist is invalid"""



from __future__ import annotations

import asyncio
import contextlib
import datetime
import typing
from typing import Any

import aiohttp
import ujson
from dacite import from_dict
from packaging.version import Version

from pylav.constants.builtin_nodes import PYLAV_NODES
from pylav.events.node import WebSocketClosedEvent
from pylav.events.plugins import SegmentSkippedEvent, SegmentsLoadedEvent
from pylav.events.track import TrackEndEvent, TrackExceptionEvent, TrackStartEvent, TrackStuckEvent
from pylav.events.track.track_start import (
    TrackStartAppleMusicEvent,
    TrackStartBandcampEvent,
    TrackStartClypitEvent,
    TrackStartDeezerEvent,
    TrackStartGCTTSEvent,
    TrackStartGetYarnEvent,
    TrackStartHTTPEvent,
    TrackStartLocalFileEvent,
    TrackStartMixCloudEvent,
    TrackStartNicoNicoEvent,
    TrackStartOCRMixEvent,
    TrackStartPornHubEvent,
    TrackStartRedditEvent,
    TrackStartSoundCloudEvent,
    TrackStartSoundgasmEvent,
    TrackStartSpeakEvent,
    TrackStartSpotifyEvent,
    TrackStartTikTokEvent,
    TrackStartTwitchEvent,
    TrackStartVimeoEvent,
    TrackStartYandexMusicEvent,
    TrackStartYouTubeEvent,
    TrackStartYouTubeMusicEvent,
)
from pylav.exceptions.node import WebsocketNotConnectedException
from pylav.exceptions.request import HTTPException
from pylav.helpers.misc import ExponentialBackoffWithReset
from pylav.helpers.time import get_now_utc
from pylav.logging import getLogger
from pylav.nodes.api.responses.plugins import SegmentSkipped, SegmentsLoaded
from pylav.nodes.api.responses.websocket import (
    Closed,
    PlayerUpdate,
    Ready,
    Stats,
    TrackEnd,
    TrackException,
    TrackStart,
    TrackStuck,
)
from pylav.nodes.utils import Stats as NodeStats
from pylav.players.query.obj import Query
from pylav.players.tracks.obj import Track
from pylav.type_hints.dict_typing import JSON_DICT_TYPE
from pylav.utils.location import get_closest_discord_region
from pylav.utils.vendor.redbot import AsyncIter

if typing.TYPE_CHECKING:
    from pylav.core.client import Client
    from pylav.nodes.node import Node
    from pylav.players.player import Player


class WebSocket:
    """Represents the WebSocket connection with Lavalink"""

    __slots__ = (
        "_node",
        "_session",
        "_ws",
        "_message_queue",
        "_host",
        "_port",
        "_password",
        "_ssl",
        "_max_reconnect_attempts",
        "_resume_timeout",
        "_resuming_configured",
        "_closers",
        "_client",
        "ready",
        "_connect_task",
        "_manual_shutdown",
        "_session_id",
        "_resumed",
        "_api_version",
        "_connecting",
        "_player_reconnect_tasks",
        "_logger",
    )

    def __init__(
        self,
        *,
        node: Node,
        host: str,
        port: int,
        password: str,
        resume_timeout: int,
        reconnect_attempts: int,
        ssl: bool,
    ):
        self._node = node
        self._logger = getLogger(f"PyLav.WebSocket-{self.node.name}")
        self._client = self._node.node_manager.client

        self._session = aiohttp.ClientSession(timeout=aiohttp.ClientTimeout(total=120), json_serialize=ujson.dumps)
        self._ws = None
        self._message_queue = []
        self._host = host
        self._port = port
        self._password = password
        self._ssl = ssl
        self._max_reconnect_attempts = reconnect_attempts

        self._resume_timeout = resume_timeout
        self._resuming_configured = False
        self._session_id: str | None = None
        self._resumed: bool | None = None
        self._api_version: int | None = None

        self._closers = (
            aiohttp.WSMsgType.CLOSE,
            aiohttp.WSMsgType.CLOSING,
            aiohttp.WSMsgType.CLOSED,
        )
        self.ready = asyncio.Event()
        self._connect_task = asyncio.ensure_future(self.connect())
        self._connect_task.add_done_callback(self._done_callback)
        self._manual_shutdown = False
        self._connecting = False
        self._player_reconnect_tasks: dict[int : asyncio.Task] = {}

    def _done_callback(self, task: asyncio.Task) -> None:
        with contextlib.suppress(asyncio.CancelledError):
            exc = task.exception()
            if exc is not None:
                self._logger.error("Error in connect task", exc_info=exc)

    @property
    def is_ready(self) -> bool:
        """Returns whether the websocket is ready"""
        return self.ready.is_set() and self.connected

    @property
    def socket_protocol(self) -> str:
        """The protocol used for the socket connection"""
        return "wss" if self._ssl else "ws"

    @property
    def lib_version(self) -> Version:
        """Returns the PyLav library version"""
        return self._client.lib_version

    @property
    def bot_id(self) -> str:
        """Returns the bot's ID"""
        return self._client.bot_id

    @property
    def node(self) -> Node:
        """Returns the :class:`Node` instance"""
        return self._node

    @property
    def client(self) -> Client:
        """Returns the :class:`Client` instance"""
        return self._client

    @property
    def connected(self) -> bool:
        """Returns whether the websocket is connected to Lavalink"""
        return self._ws is not None and not self._ws.closed and self.ready.is_set()

    @property
    def connecting(self) -> bool:
        """Returns whether the websocket is connecting to Lavalink"""
        return not self.ready.is_set()

    @property
    def session_id(self) -> str:
        """Returns the session ID"""
        return self._session_id

    async def ping(self) -> None:
        """Pings the websocket"""
        if self.connected:
            await self._ws.ping()
        else:
            raise WebsocketNotConnectedException

    async def wait_until_ready(self, timeout: float | None = None) -> None:
        await asyncio.wait_for(self.ready.wait(), timeout=timeout)

    async def configure_resume_and_timeout(self) -> None:
        payload = {"timeout": self._resume_timeout} if self._resume_timeout > 0 else {}
        payload["resuming"] = self._session_id is not None
        await self.node.patch_session(payload=payload)
        self._resuming_configured = self._session_id is not None
        self._logger.info("Node resume has been configured with sessionId: %s", self._session_id)

    async def connect(self) -> None:  # sourcery skip: low-code-quality
        """Attempts to establish a connection to Lavalink"""
        try:
            self.ready.clear()
            self.node._ready.clear()
            if self.client.is_shutting_down:
                return
            self._connecting = True
            headers = {
                "Authorization": self._password,
                "User-Id": str(self.bot_id),
                "Client-Name": f"PyLav/{self.lib_version}",
            }
            if self._node.identifier in PYLAV_NODES:
                # Since these nodes are proxied by Cloudflare - lets add a special case to properly identify them.
                self._node._region, self._node._coordinates = PYLAV_NODES[self._node.identifier]
            else:
                self._node._region, self._node._coordinates = await get_closest_discord_region(self._host)

            is_finite_retry = self._max_reconnect_attempts != -1
            max_attempts_str = self._max_reconnect_attempts if is_finite_retry else "inf"
            attempt = 0
            backoff = ExponentialBackoffWithReset(base=3)
            while not self.connected and (not is_finite_retry or attempt < self._max_reconnect_attempts):
                if self._manual_shutdown:
                    self._connecting = False
                    return
                attempt += 1
                self._logger.info(
                    "Attempting to establish WebSocket connection (%s/%s)",
                    attempt,
                    max_attempts_str,
                )
                await self.node.fetch_api_version()
                self._api_version = self.node.api_version
                if self._api_version < 4:
                    self._logger.critical(
                        "Node %s is not running a supported Lavalink version (%s:%s - %s)",
                        self._node.identifier,
                        self._host,
                        self._port,
                        self.node.version,
                    )
                    raise OSError
                if self._session_id is not None:
                    headers["Session-Id"] = self._session_id
                ws_uri = self.node.get_endpoint_websocket()
                try:
                    self._ws = await self._session.ws_connect(url=ws_uri, headers=headers, heartbeat=60, timeout=600)
                    await self._node.update_features()
                    self._connecting = False
                    backoff.reset()
                except (
                    aiohttp.ClientConnectorError,
                    aiohttp.WSServerHandshakeError,
                    aiohttp.ServerDisconnectedError,
                ) as ce:
                    if self.client.is_shutting_down:
                        self._connecting = False
                        return
                    if isinstance(ce, aiohttp.ClientConnectorError):
                        self._logger.warning(
                            "Invalid response received; this may indicate that "
                            "Lavalink is not running, or is running on a port different "
                            "to the one you passed to `add_node` (%s - %s)",
                            ws_uri,
                            headers,
                        )
                    elif isinstance(ce, aiohttp.WSServerHandshakeError):
                        if ce.status in (
                            401,
                            403,
                        ):  # Special handling for 401/403 (Unauthorized/Forbidden).
                            self._logger.warning(
                                "Authentication failed while trying to establish a connection to the node",
                            )
                            # We shouldn't try to establish any more connections as correcting this particular error
                            # would require the cog to be reloaded (or the bot to be rebooted), so further attempts
                            # would be futile, and a waste of resources.
                            self._connecting = False
                            return

                        self._logger.warning(
                            "The remote server returned code %s, "
                            "the expected code was 101. This usually "
                            "indicates that the remote server is a webserver "
                            "and not Lavalink. Check your ports, and try again",
                            ce.status,
                        )
                    await asyncio.sleep(backoff.delay())
                else:
                    #  asyncio.ensure_future(self._listen())

                    await self.node.node_manager.node_connect(self.node)
                    if self._message_queue:
                        async for message in AsyncIter(self._message_queue.copy()):
                            await self.send(**message)

                        self._message_queue.clear()
                    await self._listen()
                    self._logger.debug("Websocket lister returned")
                    # Ensure this loop doesn't proceed if _listen returns control back to this
                    # function.
                    return

            self._logger.warning(
                "A WebSocket connection could not be established within %s attempts",
                attempt,
            )
        except OSError:
            self._connecting = False
            if self._ws and not self._ws.closed and not self._ws._closing:
                await self._ws.close(code=4014, message=b"Shutting down")
            await self._session.close()
            self.ready.set()
            self.node._ready.set()
            self._connect_task.cancel()
            return
        except Exception:  # noqa
            self._logger.exception(
                "An exception occurred while attempting to connect to the node",
            )

    async def _listen(self) -> None:
        """Listens for websocket messages"""
        try:
            async for msg in self._ws:
                if self._manual_shutdown:
                    return
                self._logger.trace("Received WebSocket message: %s", msg.data)
                if msg.type == aiohttp.WSMsgType.CLOSED:
                    self._logger.info(
                        "Received close frame with code %s",
                        msg.data,
                    )
                    await self._websocket_closed(msg.data, msg.extra)
                    return
                else:
                    await self.handle_message(msg.json(loads=ujson.loads))
                # elif msg.type == aiohttp.WSMsgType.ERROR and not self.client.is_shutting_down:
                #     exc = self._ws.exception()
                #     self._logger.error("Exception in WebSocket! %s", exc)
                #     break
            await self._websocket_closed()
        except Exception:  # noqa
            if not self.client.is_shutting_down:
                self._logger.exception("Exception in WebSocket!")
                await self._websocket_closed()

    async def _websocket_closed(self, code: int = None, reason: str = None) -> None:
        """
        Handles when the websocket is closed.

        Parameters
        ----------
        code: :class:`int`
            The response code.
        reason: :class:`str`
            Reason why the websocket was closed. Defaults to `None`
        """
        self._logger.info(
            "WebSocket disconnected with the following: code=%s reason=%s",
            code,
            reason,
        )
        self._ws = None
        await self.node.node_manager.node_disconnect(self.node, code, reason)
        if not self._connect_task.cancelled():
            self._connect_task.cancel()
        if self._manual_shutdown:
            await self.close()
            return
        self._connect_task = asyncio.ensure_future(self.connect())
        self._connect_task.add_done_callback(self._done_callback)

    async def handle_message(self, data: JSON_DICT_TYPE) -> None:
        """
        Handles the response from the websocket.

        Parameters
        ----------
        data: LavalinkPlayerUpdateT|LavalinkEventT| LavalinkStatsT| LavalinkReadyT
            The data given from Lavalink.
        """
        match data["op"]:
            case "playerUpdate":
                data = from_dict(data_class=PlayerUpdate, data=data)
                await self.handle_player_update(data)
            case "stats":
                data = from_dict(data_class=Stats, data=data)
                await self.handle_stats(data)
            case "event":
                match data["type"]:
                    case "TrackStartEvent":
                        data = from_dict(data_class=TrackStart, data=data)
                    case "TrackEndEvent":
                        data = from_dict(data_class=TrackEnd, data=data)
                    case "TrackExceptionEvent":
                        data = from_dict(data_class=TrackException, data=data)
                    case "TrackStuckEvent":
                        data = from_dict(data_class=TrackStuck, data=data)
                    case "WebSocketClosedEvent":
                        data = from_dict(data_class=Closed, data=data)
                    case "SegmentsLoadedEvent":
                        data = from_dict(data_class=SegmentsLoaded, data=data)
                    case "SegmentSkippedEvent":
                        data = from_dict(data_class=SegmentSkipped, data=data)
                    case __:
                        self._logger.warning("Received unknown event: %s", data["type"])
                await self.handle_event(data)
            case "ready":
                data = from_dict(data_class=Ready, data=data)
                await self.handle_ready(data)
            case __:
                self._logger.warning("Received unknown op: %s", data["op"])

    async def handle_stats(self, data: Stats) -> None:
        """
        Handles the stats message from the websocket.

        Parameters
        ----------
        data: LavalinkStatsOpObject
            The data given from Lavalink.
        """

        self.node.stats = NodeStats(self.node, data)

    async def handle_player_update(self, data: PlayerUpdate) -> None:
        """
        Handles the player update message  from the websocket.

        Parameters
        ----------
        data: LavalinkPlayerUpdateT
            The data given from Lavalink.
        """

        if player := self.client.player_manager.get(int(data.guildId)):
            if (
                (not data.state.connected)
                and player.is_playing
                and self.ready.is_set()
                and player.connected_at < get_now_utc() - datetime.timedelta(minutes=5)
            ):
                if player.guild.id in self._player_reconnect_tasks:
                    self._player_reconnect_tasks[player.guild.id].cancel()
                self._player_reconnect_tasks[player.guild.id] = asyncio.create_task(self.maybe_reconnect_player(player))
                return
            await player._update_state(data.state)
        else:
            return

    async def maybe_reconnect_player(self, player: Player) -> None:
        """
        Attempts to reconnect the player if it is not connected.

        Parameters
        ----------
        player: :class:`Player`
            The player to reconnect.
        """
        await asyncio.sleep(5)
        session = await player.node.fetch_session_player(player.guild.id)
        if isinstance(session, HTTPException):
            return

        if (
            (not session.voice.connected)
            and player.is_playing
            and self.ready.is_set()
            and player.connected_at < get_now_utc() - datetime.timedelta(minutes=5)
        ):
            self._logger.debug("Reconnecting stalled player for %s", player.guild.id)
            await player.reconnect()
            return

    async def handle_ready(self, data: Ready) -> None:
        """
        Handles the ready message from the websocket.

        Parameters
        ----------
        data: LavalinkReadyT
            The data given from Lavalink.
        """
        self._session_id = data.sessionId
        self._resumed = data.resumed
        self.ready.set()
        self.node._ready.set()
        self._logger.info(
            "Node connected successfully and is now ready to accept commands: Session ID: %s",
            self._session_id,
        )
        await self.configure_resume_and_timeout()

    async def handle_event(
        self,
        data: (TrackStart | TrackEnd | TrackException | TrackStuck | Closed | SegmentsLoaded | SegmentSkipped),
    ):
        """
        Handles the event message from Lavalink.

        Parameters
        ----------
        data: LavalinkEventT
            The data given from Lavalink.
        """
        if self.client.is_shutting_down:
            return
        player = self.client.player_manager.get(int(data.guildId))
        if not player:
            await asyncio.sleep(3)
            player = self.client.player_manager.get(int(data.guildId))
        if not player:
            self._logger.debug(
                "Received event for non-existent player! Guild ID: %s",
                data.guildId,
            )
            return

        match data.type:
            case "TrackEndEvent":
                data = typing.cast(TrackEnd, data)

                requester = None
                track = None
                if player.current and player.current.encoded == data.encodedTrack:
                    player.current.timestamp = 0
                    requester = player.current.requester
                    track = player.current

                event = TrackEndEvent(
                    player,
                    track
                    or await Track.build_track(
                        data=data.encodedTrack,
                        requester=requester.id if requester else self._client.bot.user.id,
                        query=await Query.from_base64(data.encodedTrack),
                        node=self.node,
                    ),
                    self.node,
                    event_object=data,
                )
                await player._handle_event(event)
            case "TrackExceptionEvent":
                if self.node.identifier == player.node.identifier:
                    data = typing.cast(TrackException, data)
                    event = TrackExceptionEvent(player, player.current, node=self.node, event_object=data)
                    await player._handle_event(event)
                    self.client.dispatch_event(event)
                return
            case "TrackStartEvent":
                data = typing.cast(TrackStart, data)
                track = player.current
                event = TrackStartEvent(player, track, self.node, event_object=data)
                await self._process_track_event(player, track, self.node, data)
            case "TrackStuckEvent":
                data = typing.cast(TrackStuck, data)
                event = TrackStuckEvent(player, player.current, self.node, event_object=data)
                await player._handle_event(event)
            case "WebSocketClosedEvent":
                data = typing.cast(Closed, data)
                event = WebSocketClosedEvent(player, self.node, player.channel, event_object=data)
                if self.node.identifier == player.node.identifier:
                    await player._handle_event(event)
            case "SegmentsLoaded":
                data = typing.cast(SegmentsLoaded, data)
                event = SegmentsLoadedEvent(player, self.node, event_object=data)
            case "SegmentSkipped":
                data = typing.cast(SegmentSkipped, data)
                event = SegmentSkippedEvent(player, node=self.node, event_object=data)
            case __:
                self._logger.warning("Received unknown event: %s", data.type)
                return

        self.client.dispatch_event(event)

    async def send(self, **data: Any):
        """
        Sends a payload to Lavalink.

        Parameters
        ----------
        data: :class:`dict`
            The data sent to Lavalink.
        """
        if self._manual_shutdown:
            return
        if self.connected:
            self._logger.trace("Sending payload %s", data)
            try:
                await self._ws.send_json(data)
            except ConnectionResetError:
                self._logger.debug("Send called before WebSocket ready!")
                self._message_queue.append(data)
        else:
            self._logger.debug("Send called before WebSocket ready!")
            self._message_queue.append(data)

    async def _process_track_event(self, player: Player, track: Track, node: Node, event_object: TrackStart) -> None:
        query = await track.query()

        match query.source:
            case "YouTube Music":
                event = TrackStartYouTubeMusicEvent(player, track, node, event_object)
            case "YouTube":
                event = TrackStartYouTubeEvent(player, track, node, event_object)
            case "Spotify":
                event = TrackStartSpotifyEvent(player, track, node, event_object)
            case "Deezer":
                event = TrackStartDeezerEvent(player, track, node, event_object)
            case "Apple Music":
                event = TrackStartAppleMusicEvent(player, track, node, event_object)
            case "HTTP":
                event = TrackStartHTTPEvent(player, track, node, event_object)
            case "SoundCloud":
                event = TrackStartSoundCloudEvent(player, track, node, event_object)
            case "Clyp.it":
                event = TrackStartClypitEvent(player, track, node, event_object)
            case "Twitch":
                event = TrackStartTwitchEvent(player, track, node, event_object)
            case "Bandcamp":
                event = TrackStartBandcampEvent(player, track, node, event_object)
            case "Vimeo":
                event = TrackStartVimeoEvent(player, track, node, event_object)
            case "speak":
                event = TrackStartSpeakEvent(player, track, node, event_object)
            case "GetYarn":
                event = TrackStartGetYarnEvent(player, track, node, event_object)
            case "Mixcloud":
                event = TrackStartMixCloudEvent(player, track, node, event_object)
            case "OverClocked ReMix":
                event = TrackStartOCRMixEvent(player, track, node, event_object)
            case "Pornhub":
                event = TrackStartPornHubEvent(player, track, node, event_object)
            case "Reddit":
                event = TrackStartRedditEvent(player, track, node, event_object)
            case "SoundGasm":
                event = TrackStartSoundgasmEvent(player, track, node, event_object)
            case "TikTok":
                event = TrackStartTikTokEvent(player, track, node, event_object)
            case "Google TTS":
                event = TrackStartGCTTSEvent(player, track, node, event_object)
            case "Niconico":
                event = TrackStartNicoNicoEvent(player, track, node, event_object)
            case "Yandex Music":
                event = TrackStartYandexMusicEvent(player, track, node, event_object)
            case __:
                if query.source == "Local Files" or (
                    query._special_local and (query.is_m3u or query.is_pls or query.is_pylav)
                ):
                    event = TrackStartLocalFileEvent(player, track, node, event_object)
                else:
                    event = TrackStartEvent(player, track, node, event_object)
        self.client.dispatch_event(event)

    async def close(self) -> None:
        self._connect_task.cancel()
        if self._ws and not self._ws.closed and not self._ws._closing:
            await self._ws.close(code=4014, message=b"Shutting down")
        await self._session.close()

    async def manual_closure(self, managed_node: bool = False) -> None:
        self._manual_shutdown = managed_node
        if self._ws and not self._ws.closed and not self._ws._closing:
            with contextlib.suppress(Exception):
                await self._ws.close(code=4014, message=b"Shutting down")
        await self._websocket_closed(202, "Manual websocket shutdown requested")



from __future__ import annotations

import dataclasses
from typing import Annotated

from pylav.nodes.api.responses.filters.misc import ValueRange


@dataclasses.dataclass(repr=True, frozen=True, kw_only=True, slots=True)
class Karaoke:
    level: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None
    monoLevel: Annotated[float | None, ValueRange(min=0.0, max=1.0)] = None
    filterBand: float | None = None
    filterWidth: float | None = None

    def to_dict(self) -> dict[str, float]:
        response = {}
        if self.level is not None:
            response["level"] = self.level
        if self.monoLevel is not None:
            response["monoLevel"] = self.monoLevel
        if self.filterBand is not None:
            response["filterBand"] = self.filterBand
        if self.filterWidth is not None:
            response["filterWidth"] = self.filterWidth
        return response



from __future__ import annotations

import contextlib
import json
from typing import TYPE_CHECKING

import asyncpg
import ujson
from asyncpg import Connection
from packaging.version import parse

from pylav.constants.playlists import BUNDLED_PLAYLIST_IDS
from pylav.constants.versions import VERSION_1_0_0_0
from pylav.players.tracks.decoder import async_decoder
from pylav.storage.database.tables.config import LibConfigRow
from pylav.storage.database.tables.misc import DATABASE_ENGINE
from pylav.storage.database.tables.nodes import NodeRow
from pylav.storage.database.tables.players import PlayerRow
from pylav.storage.database.tables.playlists import PlaylistRow
from pylav.storage.database.tables.queries import QueryRow
from pylav.storage.database.tables.tracks import TrackRow
from pylav.storage.migrations.logging import LOGGER
from pylav.utils.vendor.redbot import AsyncIter

if TYPE_CHECKING:
    from pylav.storage.controllers.config import ConfigController


async def run_playlist_migration_v_1_0_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Runs playlist migration 1000.
    """
    HAS_COLUMN = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='playlist' AND column_name='tracks')
    """
    has_playlist_tracks = await connection.fetchval(HAS_COLUMN)
    if has_playlist_tracks:
        data = await connection.fetch("SELECT * FROM playlist;")
        await connection.execute("DROP TABLE playlist;")
        return data


async def run_query_migration_v_1_0_0_0(connection: Connection) -> list[asyncpg.Record] | None:
    """
    Runs playlist migration 1000.
    """
    HAS_COLUMN = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='query' AND column_name='tracks')
    """
    has_query_tracks = await connection.fetchval(HAS_COLUMN)
    if has_query_tracks:
        data = await connection.fetch("SELECT * FROM query;")
        await connection.execute("DROP TABLE query;")
        return data


async def run_player_config_v_1_0_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Migrates player config.
    """
    has_column = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='version' AND column_name='version')
    """
    has_version_column = await connection.fetchval(has_column)
    if not has_version_column:
        return []
    version = await connection.fetchval("SELECT version from version;")
    if version is None:
        return []

    version = parse(version)
    if (not version) or version < VERSION_1_0_0_0:
        return await connection.fetch("SELECT * FROM player;")
    return []


async def run_node_config_v_1_0_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Migrates player config.
    """
    has_column = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='version' AND column_name='version')
    """
    has_version_column = await connection.fetchval(has_column)
    if not has_version_column:
        return []
    version = await connection.fetchval("SELECT version from version;")
    if version is None:
        return []

    version = parse(version)
    if (not version) or version < VERSION_1_0_0_0:
        return await connection.fetch("SELECT * FROM node;")
    return []


async def run_lib_config_v_1_0_0_0(connection: Connection) -> list[asyncpg.Record]:
    """
    Migrates player config.
    """
    has_column = """
    SELECT EXISTS (SELECT 1
    FROM information_schema.columns
    WHERE table_name='version' AND column_name='version')
    """
    has_version_column = await connection.fetchval(has_column)
    if not has_version_column:
        return []
    version = await connection.fetchval("SELECT version from version;")
    if version is None:
        return []

    version = parse(version)
    if (not version) or version < VERSION_1_0_0_0:
        return await connection.fetch("SELECT * FROM lib_config;")


async def migrate_playlists_v_1_0_0_0(playlists: list[asyncpg.Record]) -> None:
    for playlist in playlists:
        if playlist["id"] in BUNDLED_PLAYLIST_IDS:
            continue
        defaults = {
            PlaylistRow.name: playlist["name"],
            PlaylistRow.scope: playlist["scope"],
            PlaylistRow.author: playlist["author"],
            PlaylistRow.url: playlist["url"],
        }
        playlist_row = await PlaylistRow.objects().get_or_create(PlaylistRow.id == playlist["id"], defaults)
        # noinspection PyProtectedMember
        if not playlist_row._was_created:
            await PlaylistRow.update(defaults).where(PlaylistRow.id == playlist["id"])
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        tracks = json.loads(playlist["tracks"]) if playlist["tracks"] else []
        async for track in AsyncIter(tracks):
            with contextlib.suppress(Exception):
                data = await async_decoder(track)  # TODO: Make an API call to the public node?
                new_tracks.append(await TrackRow.get_or_create(data.encoded, data.info.to_database()))
        if new_tracks:
            await playlist_row.add_m2m(*new_tracks, m2m=PlaylistRow.tracks)


async def migrate_queries_v_1_0_0_0(queries: list[asyncpg.Record]) -> None:

    for query in queries:
        defaults = {QueryRow.name: query["name"]}
        query_row = await QueryRow.objects().get_or_create(QueryRow.identifier == query["identifier"], defaults)

        # noinspection PyProtectedMember
        if not query_row._was_created:
            await QueryRow.update(defaults).where(QueryRow.identifier == query["identifier"])
        new_tracks = []
        # TODO: Optimize this, after https://github.com/piccolo-orm/piccolo/discussions/683 is answered or fixed
        tracks = json.loads(query["tracks"]) if query["tracks"] else []
        async for track in AsyncIter(tracks):
            with contextlib.suppress(Exception):
                data = await async_decoder(track)  # TODO: Make an API call to the public node?
                new_tracks.append(await TrackRow.get_or_create(data.encoded, data.info.to_database()))
        if new_tracks:
            await query_row.add_m2m(*new_tracks, m2m=QueryRow.tracks)


async def migrate_player_config_v_1_0_0_0(players: list[asyncpg.Record]) -> None:
    bulk_insert = []
    for player in players:
        data = {
            "id": player["id"],
            "bot": player["bot"],
            "volume": player["volume"],
            "max_volume": player["max_volume"],
            "auto_play_playlist_id": player["auto_play_playlist_id"],
            "text_channel_id": player["text_channel_id"],
            "notify_channel_id": player["notify_channel_id"],
            "forced_channel_id": player["forced_channel_id"],
            "repeat_current": player["repeat_current"],
            "repeat_queue": player["repeat_queue"],
            "shuffle": player["shuffle"],
            "auto_shuffle": player["auto_shuffle"],
            "auto_play": player["auto_play"],
            "self_deaf": player["self_deaf"],
            "empty_queue_dc": ujson.loads(player["empty_queue_dc"]),
            "alone_dc": ujson.loads(player["alone_dc"]),
            "alone_pause": ujson.loads(player["alone_pause"]),
            "extras": ujson.loads(player["extras"]),
            "effects": ujson.loads(player["effects"]),
            "dj_users": player["dj_users"],
            "dj_roles": player["dj_roles"],
        }

        if player["id"] == 0:
            data = {
                PlayerRow.volume: player["volume"],
                PlayerRow.max_volume: player["max_volume"],
                PlayerRow.auto_play_playlist_id: player["auto_play_playlist_id"],
                PlayerRow.text_channel_id: player["text_channel_id"],
                PlayerRow.notify_channel_id: player["notify_channel_id"],
                PlayerRow.forced_channel_id: player["forced_channel_id"],
                PlayerRow.repeat_current: player["repeat_current"],
                PlayerRow.repeat_queue: player["repeat_queue"],
                PlayerRow.shuffle: player["shuffle"],
                PlayerRow.auto_shuffle: player["auto_shuffle"],
                PlayerRow.auto_play: player["auto_play"],
                PlayerRow.self_deaf: player["self_deaf"],
                PlayerRow.empty_queue_dc: ujson.loads(player["empty_queue_dc"]),
                PlayerRow.alone_dc: ujson.loads(player["alone_dc"]),
                PlayerRow.alone_pause: ujson.loads(player["alone_pause"]),
                PlayerRow.extras: ujson.loads(player["extras"]),
                PlayerRow.effects: ujson.loads(player["effects"]),
                PlayerRow.dj_users: player["dj_users"],
                PlayerRow.dj_roles: player["dj_roles"],
            }

            playerobj = await PlayerRow.objects().get_or_create(
                (PlayerRow.id == player["id"]) & (PlayerRow.bot == player["bot"]), defaults=data
            )
            if not playerobj._was_created:
                await PlayerRow.update(data).where((PlayerRow.id == player["id"]) & (PlayerRow.bot == player["bot"]))

        else:
            bulk_insert.append(PlayerRow(**data))
    if bulk_insert:
        await PlayerRow.insert(*bulk_insert)


async def migrate_node_config_v_1_0_0_0(nodes: list[asyncpg.Record]) -> None:
    for node in nodes:
        data = {
            NodeRow.name: node["name"],
            NodeRow.ssl: node["ssl"],
            NodeRow.resume_timeout: node["resume_timeout"],
            NodeRow.reconnect_attempts: node["reconnect_attempts"],
            NodeRow.search_only: node["search_only"],
            NodeRow.managed: node["managed"],
            NodeRow.disabled_sources: node["disabled_sources"],
            NodeRow.extras: ujson.loads(node["extras"]),
            NodeRow.yaml: ujson.loads(node["yaml"]),
        }
        node_obj = await NodeRow.objects().get_or_create(NodeRow.id == node["id"], defaults=data)
        if not node_obj._was_created:
            await NodeRow.update(data).where(NodeRow.id == node["id"])


async def migrate_lib_config_v_1_0_0_0(configs: list[asyncpg.Record]) -> None:
    for config in configs:
        data = {
            LibConfigRow.config_folder: config["config_folder"],
            LibConfigRow.java_path: config["java_path"],
            LibConfigRow.enable_managed_node: config["enable_managed_node"],
            LibConfigRow.auto_update_managed_nodes: config["auto_update_managed_nodes"],
            LibConfigRow.localtrack_folder: config["localtrack_folder"],
            LibConfigRow.download_id: config["download_id"],
            LibConfigRow.update_bot_activity: config["update_bot_activity"],
            LibConfigRow.use_bundled_pylav_external: config["use_bundled_pylav_external"],
            LibConfigRow.use_bundled_lava_link_external: False,
            LibConfigRow.extras: ujson.loads(config["extras"]),
            LibConfigRow.next_execution_update_bundled_playlists: config["next_execution_update_bundled_playlists"],
            LibConfigRow.next_execution_update_bundled_external_playlists: config[
                "next_execution_update_bundled_external_playlists"
            ],
            LibConfigRow.next_execution_update_external_playlists: config["next_execution_update_external_playlists"],
        }
        config_obj = await LibConfigRow.objects().get_or_create(
            (LibConfigRow.id == config["id"]) & (LibConfigRow.bot == config["bot"]), defaults=data
        )
        if not config_obj._was_created:
            await LibConfigRow.update(data).where(
                (LibConfigRow.id == config["id"]) & (LibConfigRow.bot == config["bot"])
            )


async def run_low_level_migrations(migrator: ConfigController) -> dict[str, dict[str, list[asyncpg.Record] | None]]:
    """
    Runs migrations.
    """
    migration_data = {}
    con = await DATABASE_ENGINE.get_new_connection()
    await low_level_v_1_0_0_migration(con, migration_data, migrator)
    return migration_data


async def low_level_v_1_0_0_migration(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], migrator: ConfigController
) -> None:
    version = "1.0.0"
    migration_data[version] = {}
    await low_level_v_1_0_0_playlists(con, migration_data, version)
    await low_level_v_1_0_0_queries(con, migration_data, version)
    await low_level_v_1_0_0_players(con, migration_data, version)
    await low_level_v_1_0_0_lib(con, migration_data, version)
    await low_level_v_1_0_0_nodes(con, migration_data, version)
    if migration_data[version]:
        await migrator.reset_database()


async def low_level_v_1_0_0_nodes(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    node_config_data_1000 = await run_node_config_v_1_0_0_0(con)
    if node_config_data_1000:
        migration_data[version]["node"] = node_config_data_1000


async def low_level_v_1_0_0_lib(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    lib_config_data_1000 = await run_lib_config_v_1_0_0_0(con)
    if lib_config_data_1000:
        migration_data[version]["lib"] = lib_config_data_1000


async def low_level_v_1_0_0_players(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    player_data_1000 = await run_player_config_v_1_0_0_0(con)
    if player_data_1000:
        migration_data[version]["player"] = player_data_1000


async def low_level_v_1_0_0_queries(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    query_data_1000 = await run_query_migration_v_1_0_0_0(con)
    if query_data_1000:
        migration_data[version]["query"] = query_data_1000


async def low_level_v_1_0_0_playlists(
    con: Connection, migration_data: dict[str, dict[str, list[asyncpg.Record]] | None], version: str
) -> None:
    playlist_data_1000 = await run_playlist_migration_v_1_0_0_0(con)
    if playlist_data_1000:
        migration_data[version]["playlist"] = playlist_data_1000


async def migrate_data(data: dict[str, dict[str, list[asyncpg.Record]]]) -> None:
    """
    Migrates data.
    """

    for version, migrations in data.items():
        match version:
            case "1.0.0":
                if "playlist" in migrations and migrations["playlist"]:
                    LOGGER.info("-----------Migrating Playlist data to PyLav 1.0.0 ---------")
                    await migrate_playlists_v_1_0_0_0(migrations["playlist"])
                if "query" in migrations and migrations["query"]:
                    LOGGER.info("-----------Migrating Query data to PyLav 1.0.0 ---------")
                    await migrate_queries_v_1_0_0_0(migrations["query"])
                if "player" in migrations and migrations["player"]:
                    LOGGER.info("-----------Migrating Player Config to PyLav 1.0.0 ---------")
                    await migrate_player_config_v_1_0_0_0(migrations["player"])
                if "lib" in migrations and migrations["lib"]:
                    LOGGER.info("-----------Migrating Lib Config to PyLav 1.0.0 ---------")
                    await migrate_lib_config_v_1_0_0_0(migrations["lib"])
                if "node" in migrations and migrations["node"]:
                    LOGGER.info("-----------Migrating Node Config to PyLav 1.0.0 ---------")
                    await migrate_node_config_v_1_0_0_0(migrations["node"])



from __future__ import annotations

import datetime
import itertools
import re
from collections.abc import Callable
from typing import Any
from urllib.parse import urljoin as _urljoin

import iso8601

from pylav.extension.m3u import protocols
from pylav.type_hints.generics import ANY_GENERIC_TYPE

# coding: utf-8
# Copyright 2014 Globo.com Player authors. All rights reserved.
# Use of this source code is governed by a MIT License
# license that can be found in the LICENSE file.


"""
http://tools.ietf.org/html/draft-pantos-http-live-streaming-08#section-3.2
http://stackoverflow.com/questions/2785755/how-to-split-but-ignore-separators-in-quoted-strings-in-python
"""
ATTRIBUTELISTPATTERN = re.compile(r"""((?:[^,"']|"[^"]*"|'[^']*')+)""")
URI_PREFIXES = ("https://", "http://", "s3://", "s3a://", "s3n://")


def cast_date_time(value: str) -> datetime.datetime:
    return iso8601.parse_date(value)


def format_date_time(value: datetime.datetime) -> str:
    return value.isoformat()


class ParseError(Exception):
    def __init__(self, lineno: int, line: str) -> None:
        self.lineno = lineno
        self.line = line

    def __str__(self) -> str:
        return f"Syntax error in manifest on line {self.lineno:d}: {self.line}"


def parse(content: str, strict: bool = False, custom_tags_parser: bool = None) -> dict[str, Any]:
    """
    Given a M3U8 playlist content returns a dictionary with all data found
    """
    data = {
        "media_sequence": 0,
        "is_variant": False,
        "is_endlist": False,
        "is_i_frames_only": False,
        "is_independent_segments": False,
        "playlist_type": None,
        "playlists": [],
        "segments": [],
        "iframe_playlists": [],
        "media": [],
        "keys": [],
        "rendition_reports": [],
        "skip": {},
        "part_inf": {},
        "session_data": [],
        "session_keys": [],
    }

    state = {
        "expect_segment": False,
        "expect_playlist": False,
        "current_key": None,
        "current_segment_map": None,
    }

    lineno = 0
    for line in string_to_lines(content):
        lineno += 1
        line = line.strip()

        # Call custom parser if needed
        if line.startswith("#") and callable(custom_tags_parser):
            go_to_next_line = custom_tags_parser(line, lineno, data, state)

            # Do not try to parse other standard tags on this line if custom_tags_parser function returns 'True'
            if go_to_next_line:
                continue

        if line.startswith(protocols.EXT_X_BYTE_RANGE):
            _parse_byterange(line, state)
            state["expect_segment"] = True
            continue

        _process_line(content, data, line, lineno, state, strict)

    # there could be remaining partial segments
    if "segment" in state:
        data["segments"].append(state.pop("segment"))

    return data


def _process_line(content, data, line, lineno, state, strict):  # sourcery skip: low-code-quality
    if line.startswith(protocols.EXT_X_BIT_RATE):
        _parse_bitrate(line, state)

    elif line.startswith(protocols.EXT_X_TARGET_DURATION):
        _parse_simple_parameter(line, data, float)

    elif line.startswith(protocols.EXT_X_MEDIA_SEQUENCE):
        _parse_simple_parameter(line, data, int)

    elif line.startswith(protocols.EXT_X_DISCONTINUITY_SEQUENCE):
        _parse_simple_parameter(line, data, int)

    elif line.startswith(protocols.EXT_X_PROGRAM_DATE_TIME):
        _, program_date_time = _parse_simple_parameter_raw_value(line, cast_date_time)
        if not data.get("program_date_time"):
            data["program_date_time"] = program_date_time
        state["current_program_date_time"] = program_date_time
        state["program_date_time"] = program_date_time

    elif line.startswith(protocols.EXT_X_DISCONTINUITY):
        state["discontinuity"] = True

    elif line.startswith(protocols.EXT_X_CUE_OUT_CONT):
        _parse_cueout_cont(line, state)
        state["cue_out"] = True

    elif line.startswith(protocols.EXT_X_CUE_OUT):
        _parse_cueout(line, state, string_to_lines(content)[lineno - 2])
        state["cue_out_start"] = True
        state["cue_out"] = True

    elif line.startswith(protocols.EXT_X_CUE_IN):
        state["cue_in"] = True

    elif line.startswith(protocols.EXT_X_CUE_SPAN):
        state["cue_out"] = True

    elif line.startswith(protocols.EXT_X_VERSION):
        _parse_simple_parameter(line, data, int)

    elif line.startswith(protocols.EXT_X_ALLOW_CACHE):
        _parse_simple_parameter(line, data)

    elif line.startswith(protocols.EXT_X_KEY):
        key = _parse_key(line)
        state["current_key"] = key
        if key not in data["keys"]:
            data["keys"].append(key)

    elif line.startswith(protocols.EXT_INF):
        _parse_extinf(line, data, state, lineno, strict)
        state["expect_segment"] = True

    elif line.startswith(protocols.EXT_X_STREAM_INF):
        state["expect_playlist"] = True
        _parse_stream_inf(line, data, state)

    elif line.startswith(protocols.EXT_X_I_FRAME_STREAM_INF):
        _parse_i_frame_stream_inf(line, data)

    elif line.startswith(protocols.EXT_X_MEDIA):
        _parse_media(line, data, state)

    elif line.startswith(protocols.EXT_X_PLAYLIST_TYPE):
        _parse_simple_parameter(line, data)

    elif line.startswith(protocols.EXT_I_FRAMES_ONLY):
        data["is_i_frames_only"] = True

    elif line.startswith(protocols.EXT_IS_INDEPENDENT_SEGMENTS):
        data["is_independent_segments"] = True

    elif line.startswith(protocols.EXT_X_END_LIST):
        data["is_endlist"] = True

    elif line.startswith(protocols.EXT_X_MAP):
        quoted_parser = remove_quotes_parser("uri")
        segment_map_info = _parse_attribute_list(protocols.EXT_X_MAP, line, quoted_parser)
        state["current_segment_map"] = segment_map_info
        # left for backward compatibility
        data["segment_map"] = segment_map_info

    elif line.startswith(protocols.EXT_X_START):
        attribute_parser = {"time_offset": lambda x: float(x)}
        start_info = _parse_attribute_list(protocols.EXT_X_START, line, attribute_parser)
        data["start"] = start_info

    elif line.startswith(protocols.EXT_X_SERVER_CONTROL):
        _parse_server_control(line, data, state)

    elif line.startswith(protocols.EXT_X_PART_INF):
        _parse_part_inf(line, data, state)

    elif line.startswith(protocols.EXT_X_RENDITION_REPORT):
        _parse_rendition_report(line, data, state)

    elif line.startswith(protocols.EXT_X_PART):
        _parse_part(line, data, state)

    elif line.startswith(protocols.EXT_X_SKIP):
        _parse_skip(line, data, state)

    elif line.startswith(protocols.EXT_X_SESSION_DATA):
        _parse_session_data(line, data, state)

    elif line.startswith(protocols.EXT_X_SESSION_KEY):
        _parse_session_key(line, data, state)

    elif line.startswith(protocols.EXT_X_PRELOAD_HINT):
        _parse_preload_hint(line, data, state)

    elif line.startswith(protocols.EXT_X_DATERANGE):
        _parse_daterange(line, data, state)

    elif line.startswith(protocols.EXT_X_GAP):
        state["gap"] = True

    elif line.startswith(protocols.EXT_X_CONTENT_STEERING):
        _parse_content_steering(line, data, state)

    elif line.startswith(protocols.EXT_M3U):
        # We don't parse #EXTM3U, it just should to be present
        pass

    elif line.strip() == "":
        # blank lines are legal
        pass

    elif state["expect_segment"]:
        _parse_ts_chunk(line, data, state)
        state["expect_segment"] = False

    elif state["expect_playlist"]:
        _parse_variant_playlist(line, data, state)
        state["expect_playlist"] = False

    elif strict:
        raise ParseError(lineno, line)


def _parse_key(line: str) -> dict[str, Any]:
    params = ATTRIBUTELISTPATTERN.split(line.replace(f"{protocols.EXT_X_KEY}:", ""))[1::2]
    key = {}
    for param in params:
        name, value = param.split("=", 1)
        key[normalize_attribute(name)] = remove_quotes(value)
    return key


def _parse_extinf(line: str, data: dict, state: dict, lineno: int, strict: bool) -> None:  # noqa
    chunks = line.replace(f"{protocols.EXT_INF}:", "").split(",", 1)
    if len(chunks) == 2:
        duration, title = chunks
    elif len(chunks) == 1:
        if strict:
            raise ParseError(lineno, line)
        duration = chunks[0]
        title = ""
    if "segment" not in state:
        state["segment"] = {}
    state["segment"]["duration"] = float(duration)  # noqa
    state["segment"]["title"] = title  # noqa


def _parse_ts_chunk(line: str, data: dict, state: dict) -> None:
    segment = state.pop("segment")
    if state.get("program_date_time"):
        segment["program_date_time"] = state.pop("program_date_time")
    if state.get("current_program_date_time"):
        segment["current_program_date_time"] = state["current_program_date_time"]
        state["current_program_date_time"] += datetime.timedelta(seconds=segment["duration"])
    segment["uri"] = line
    segment["cue_in"] = state.pop("cue_in", False)
    segment["cue_out"] = state.pop("cue_out", False)
    segment["cue_out_start"] = state.pop("cue_out_start", False)
    scte_op = state.pop if segment["cue_in"] else state.get
    segment["scte35"] = scte_op("current_cue_out_scte35", None)
    segment["scte35_duration"] = scte_op("current_cue_out_duration", None)
    segment["discontinuity"] = state.pop("discontinuity", False)
    if state.get("current_key"):
        segment["key"] = state["current_key"]
    elif None not in data["keys"]:
        data["keys"].append(None)
    if state.get("current_segment_map"):
        segment["init_section"] = state["current_segment_map"]
    segment["dateranges"] = state.pop("dateranges", None)
    segment["gap_tag"] = state.pop("gap", None)
    data["segments"].append(segment)


def _parse_attribute_list(prefix: str, line: str, atribute_parser: dict) -> dict:
    params = ATTRIBUTELISTPATTERN.split(line.replace(f"{prefix}:", ""))[1::2]

    attributes = {}
    for param in params:
        name, value = param.split("=", 1)
        name = normalize_attribute(name)

        if name in atribute_parser:
            value = atribute_parser[name](value)

        attributes[name] = value

    return attributes


def _parse_stream_inf(line: str, data: dict, state: dict) -> None:
    data["is_variant"] = True
    data["media_sequence"] = None
    atribute_parser = remove_quotes_parser("codecs", "audio", "video", "subtitles", "closed_captions", "pathway_id")
    atribute_parser["program_id"] = int
    atribute_parser["bandwidth"] = lambda x: int(float(x))
    atribute_parser["average_bandwidth"] = int
    atribute_parser["frame_rate"] = float
    atribute_parser["video_range"] = str
    atribute_parser["hdcp_level"] = str
    state["stream_info"] = _parse_attribute_list(protocols.EXT_X_STREAM_INF, line, atribute_parser)


def _parse_i_frame_stream_inf(line: str, data: dict) -> None:
    atribute_parser = remove_quotes_parser("codecs", "uri", "pathway_id")
    atribute_parser["program_id"] = int
    atribute_parser["bandwidth"] = int
    atribute_parser["average_bandwidth"] = int
    atribute_parser["video_range"] = str
    atribute_parser["hdcp_level"] = str
    iframe_stream_info = _parse_attribute_list(protocols.EXT_X_I_FRAME_STREAM_INF, line, atribute_parser)
    iframe_playlist = {"uri": iframe_stream_info.pop("uri"), "iframe_stream_info": iframe_stream_info}

    data["iframe_playlists"].append(iframe_playlist)


def _parse_media(line: str, data: dict, state: dict) -> None:  # noqa
    quoted = remove_quotes_parser(
        "uri", "group_id", "language", "assoc_language", "name", "instream_id", "characteristics", "channels"
    )
    media = _parse_attribute_list(protocols.EXT_X_MEDIA, line, quoted)
    data["media"].append(media)


def _parse_variant_playlist(line: str, data: dict, state: dict) -> None:
    playlist = {"uri": line, "stream_info": state.pop("stream_info")}

    data["playlists"].append(playlist)


def _parse_bitrate(line: str, state: dict) -> None:
    if "segment" not in state:
        state["segment"] = {}
    state["segment"]["bitrate"] = line.replace(f"{protocols.EXT_X_BIT_RATE}:", "")


def _parse_byterange(line: str, state: dict) -> None:
    if "segment" not in state:
        state["segment"] = {}
    state["segment"]["byterange"] = line.replace(f"{protocols.EXT_X_BYTE_RANGE}:", "")


def _parse_simple_parameter_raw_value(line: str, cast_to: Callable = str, normalize=False):
    param, value = line.split(":", 1)
    param = normalize_attribute(param.replace("#EXT-X-", ""))
    if normalize:
        value = value.strip().lower()
    return param, cast_to(value)


def _parse_and_set_simple_parameter_raw_value(line: str, data: dict, cast_to: type = str, normalize=False):
    param, value = _parse_simple_parameter_raw_value(line, cast_to, normalize)
    data[param] = value
    return data[param]


def _parse_simple_parameter(line: str, data: dict, cast_to: type = str):
    return _parse_and_set_simple_parameter_raw_value(line, data, cast_to, True)


def _parse_cueout_cont(line: str, state) -> None:
    elements = line.split(":", 1)
    if len(elements) != 2:
        return
    param, value = elements
    if res := re.match(".*Duration=(.*),SCTE35=(.*)$", value):
        state["current_cue_out_duration"] = res[1]
        state["current_cue_out_scte35"] = res[2]


def _cueout_no_duration(line: str) -> None | tuple[None, None]:
    # this needs to be called first since line.split in all other
    # parsers will throw a ValueError if passed just this tag
    if line == protocols.EXT_X_CUE_OUT:
        return None, None


def _cueout_elemental(line: str, state: dict, prevline: str) -> None | tuple[str, str]:  # noqa
    param, value = line.split(":", 1)
    if res := re.match(".*EXT-OATCLS-SCTE35:(.*)$", prevline):
        return res[1], value
    else:
        return None


def _cueout_envivio(line: str, state: dict, prevline: str) -> None | tuple[str, str]:  # noqa
    param, value = line.split(":", 1)
    if res := re.match('.*DURATION=(.*),.*,CUE="(.*)"', value):
        return res[2], res[1]
    else:
        return None


def _cueout_duration(line: str) -> None | tuple[None, str]:
    # this needs to be called after _cueout_elemental
    # as it would capture those cues incompletely
    # This was added separately rather than modifying "simple"
    param, value = line.split(":", 1)
    if res := re.match(r"DURATION=(.*)", value):
        return None, res[1]


def _cueout_simple(line) -> None | tuple[None, str]:
    # this needs to be called after _cueout_elemental
    # as it would capture those cues incompletely
    param, value = line.split(":", 1)
    if res := re.match(r"^(\d+(?:\.\d)?\d*)$", value):
        return None, res[1]


def _parse_cueout(line: str, state: dict, prevline) -> None:
    if _cueout_state := (
        _cueout_no_duration(line)
        or _cueout_elemental(line, state, prevline)
        or _cueout_envivio(line, state, prevline)
        or _cueout_duration(line)
        or _cueout_simple(line)
    ):
        state["current_cue_out_scte35"] = _cueout_state[0]
        state["current_cue_out_duration"] = _cueout_state[1]


def _parse_server_control(line: str, data, state) -> None:  # noqa
    attribute_parser = {
        "can_block_reload": str,
        "hold_back": lambda x: float(x),
        "part_hold_back": lambda x: float(x),
        "can_skip_until": lambda x: float(x),
        "can_skip_dateranges": str,
    }

    data["server_control"] = _parse_attribute_list(protocols.EXT_X_SERVER_CONTROL, line, attribute_parser)


def _parse_part_inf(line: str, data: dict, state: dict) -> None:  # noqa
    attribute_parser = {"part_target": lambda x: float(x)}

    data["part_inf"] = _parse_attribute_list(protocols.EXT_X_PART_INF, line, attribute_parser)


def _parse_rendition_report(line: str, data: dict, state: dict) -> None:  # noqa
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["last_msn"] = int
    attribute_parser["last_part"] = int

    rendition_report = _parse_attribute_list(protocols.EXT_X_RENDITION_REPORT, line, attribute_parser)

    data["rendition_reports"].append(rendition_report)


def _parse_part(line: str, data: dict, state: dict) -> None:  # noqa
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["duration"] = lambda x: float(x)
    attribute_parser["independent"] = str
    attribute_parser["gap"] = str
    attribute_parser["byterange"] = str

    part = _parse_attribute_list(protocols.EXT_X_PART, line, attribute_parser)

    # this should always be true according to spec
    if state.get("current_program_date_time"):
        part["program_date_time"] = state["current_program_date_time"]
        state["current_program_date_time"] += datetime.timedelta(seconds=part["duration"])

    part["dateranges"] = state.pop("dateranges", None)
    part["gap_tag"] = state.pop("gap", None)

    if "segment" not in state:
        state["segment"] = {}
    segment = state["segment"]
    if "parts" not in segment:
        segment["parts"] = []

    segment["parts"].append(part)


def _parse_skip(line: str, data: dict, state: dict) -> None:  # noqa
    attribute_parser = remove_quotes_parser("recently_removed_dateranges")
    attribute_parser["skipped_segments"] = int

    data["skip"] = _parse_attribute_list(protocols.EXT_X_SKIP, line, attribute_parser)


def _parse_session_data(line: str, data: dict, state: dict) -> None:  # noqa
    quoted = remove_quotes_parser("data_id", "value", "uri", "language")
    session_data = _parse_attribute_list(protocols.EXT_X_SESSION_DATA, line, quoted)
    data["session_data"].append(session_data)


def _parse_session_key(line: str, data: dict, state: dict) -> None:  # noqa
    params = ATTRIBUTELISTPATTERN.split(line.replace(f"{protocols.EXT_X_SESSION_KEY}:", ""))[1::2]
    key = {}
    for param in params:
        name, value = param.split("=", 1)
        key[normalize_attribute(name)] = remove_quotes(value)
    data["session_keys"].append(key)


def _parse_preload_hint(line: str, data: dict, state: dict) -> None:  # noqa
    attribute_parser = remove_quotes_parser("uri")
    attribute_parser["type"] = str
    attribute_parser["byterange_start"] = int
    attribute_parser["byterange_length"] = int

    data["preload_hint"] = _parse_attribute_list(protocols.EXT_X_PRELOAD_HINT, line, attribute_parser)


def _parse_daterange(line: str, date: dict, state: dict) -> None:  # noqa
    attribute_parser = remove_quotes_parser("id", "class", "start_date", "end_date")
    attribute_parser["duration"] = float
    attribute_parser["planned_duration"] = float
    attribute_parser["end_on_next"] = str
    attribute_parser["scte35_cmd"] = str
    attribute_parser["scte35_out"] = str
    attribute_parser["scte35_in"] = str

    parsed = _parse_attribute_list(protocols.EXT_X_DATERANGE, line, attribute_parser)

    if "dateranges" not in state:
        state["dateranges"] = []

    state["dateranges"].append(parsed)


def _parse_content_steering(line: str, data: dict, state: dict) -> None:  # noqa
    attribute_parser = remove_quotes_parser("server_uri", "pathway_id")

    data["content_steering"] = _parse_attribute_list(protocols.EXT_X_CONTENT_STEERING, line, attribute_parser)


def string_to_lines(string: str) -> list[str]:
    return string.strip().splitlines()


def remove_quotes_parser(*attrs: str) -> dict[str, Callable]:
    return dict(zip(attrs, itertools.repeat(remove_quotes)))


def remove_quotes(string: str) -> str:
    """
    Remove quotes from string.

    Ex.:
      "foo" -> foo
      'foo' -> foo
      'foo  -> 'foo

    """
    quotes = ('"', "'")
    if string.startswith(quotes) and string.endswith(quotes):
        return string[1:-1]
    return string


def normalize_attribute(attribute: str) -> str:
    return attribute.replace("-", "_").lower().strip()


def is_url(uri: str) -> bool:
    return f"{uri}".startswith(URI_PREFIXES)


def urljoin(base: str, url: str) -> str:
    base = base.replace("://", "\1")
    url = url.replace("://", "\1")
    while "//" in base:
        base = base.replace("//", "/\0/")
    while "//" in url:
        url = url.replace("//", "/\0/")
    return _urljoin(base.replace("\1", "://"), url.replace("\1", "://")).replace("\0", "")


def get_segment_custom_value(state: dict, key: str, default: ANY_GENERIC_TYPE = None) -> ANY_GENERIC_TYPE | Any:
    """
    Helper function for getting custom values for Segment
    Are useful with custom_tags_parser
    """
    if "segment" not in state:
        return default
    if "custom_parser_values" not in state["segment"]:
        return default
    return state["segment"]["custom_parser_values"].get(key, default)


def save_segment_custom_value(state: dict, key: str, value: Any) -> None:
    """
    Helper function for saving custom values for Segment
    Are useful with custom_tags_parser
    """
    if "segment" not in state:
        state["segment"] = {}

    if "custom_parser_values" not in state["segment"]:
        state["segment"]["custom_parser_values"] = {}

    state["segment"]["custom_parser_values"][key] = value
